# -*- coding: utf-8 -*-
"""
Created on Tue Oct 25 12:31:28 2022

@author: ereznic2
"""
# Specify file path to PyFAST
import sys
#sys.path.insert(1,'../PyFAST/')
import numpy as np
import pandas as pd
# sys.path.insert(1,sys.path[0] + '/ProFAST-main/') #ESG
import ProFAST

from greenheart.to_organize.H2_Analysis import LCA_single_scenario_ProFAST

# sys.path.append('../ProFAST/')

pf = ProFAST.ProFAST()


def run_profast_for_hydrogen(hopp_dict,electrolyzer_size_mw,H2_Results,\
                            electrolyzer_system_capex_kw,user_defined_time_between_replacement,electrolyzer_energy_kWh_per_kg,hydrogen_storage_capacity_kg,hydrogen_storage_cost_USDprkg,\
                            capex_desal,opex_desal,plant_life,water_cost,wind_size_mw,solar_size_mw,storage_size_mw,renewable_plant_cost_info,wind_om_cost_kw,hybrid_plant,\
                            grid_connection_scenario, atb_year, site_name, policy_option, policy, energy_to_electrolyzer, combined_hybrid_power_production_hopp,combined_hybrid_curtailment_hopp,\
                            energy_shortfall_hopp, elec_price,grid_prices_interpolated_USDperkwh, grid_price_scenario,user_defined_stack_replacement_time,use_optimistic_pem_efficiency,wind_annual_energy_MWh,solar_annual_energy_MWh,solar_ITC): 
    # mwh_to_kwh = 0.001
    # plant_life=useful_life
    # electrolyzer_system_capex_kw = electrolyzer_capex_kw
    energy_used_per_year_kWhpryr = H2_Results['Performance Schedules']['Annual Energy Used [kWh/year]'].values
    elec_efficiency_per_yr_kWhprkg=H2_Results['Performance Schedules']['Annual Average Efficiency [kWh/kg]'].values
    # H2_Results['Performance Schedules']['Annual Average Efficiency [%-HHV]']
    h2prod_per_year_kgpryr = H2_Results['Performance Schedules']['Annual H2 Production [kg/year]'].values
    water_consumption_avg_kgH2O_prkgH2 = 10 #ESG, we should make this input 09/05
    electrolysis_plant_capacity_kgperday = 24*H2_Results['new_H2_Results']['Rated BOL: H2 Production [kg/hr]']
    # debt_equity_split = 57.983
    replace_stacks_individually = False #TODO, make input (changes refurbishment period)

    water_consumption_avg_galH2O_prkgH2 = water_consumption_avg_kgH2O_prkgH2/3.79

    # Design point electricity consumption
    if use_optimistic_pem_efficiency:
        elec_consumption_kWhprkg_design = electrolyzer_energy_kWh_per_kg
    else:
        elec_consumption_kWhprkg_design=H2_Results['new_H2_Results']['Rated BOL: Efficiency [kWh/kg]'] #new 09/05
        # elec_consumption_kWhprkg_design=H2_Results['Rated kWh/kg-H2'] #new 09/05

    # Calculate electrolyzer production capacity
    # electrolysis_plant_capacity_kgperday=   electrolyzer_size_mw/elec_consumption_kWhprkg_design*1000*24
    #electrolysis_plant_capacity_kgperday = electrolyzer_size_mw*electrolyzer_design_efficiency_HHV/h2_HHV*3600*24

    # Installed capital cost
    electrolyzer_installation_factor = 12/100  #[%] for stack cost

    # Indirect capital cost as a percentage of installed capital cost
    site_prep = 2/100   #[%]
    engineering_design = 10/100 #[%]
    project_contingency = 15/100 #[%]
    permitting = 15/100     #[%]
    land_cost = 250000   #[$]

    stack_replacement_cost = 15/100  #[% of installed capital cost]

    # Calculate electrolyzer installation cost
    total_direct_electrolyzer_cost_kw = (electrolyzer_system_capex_kw * (1+electrolyzer_installation_factor)) \

    electrolyzer_total_installed_capex = total_direct_electrolyzer_cost_kw*electrolyzer_size_mw*1000

    electrolyzer_indirect_cost = electrolyzer_total_installed_capex*(site_prep+engineering_design+project_contingency+permitting)

    compressor_capex_USDprkWe_of_electrolysis = 39

    # Calculate capital costs
    capex_electrolyzer_overnight = electrolyzer_total_installed_capex + electrolyzer_indirect_cost
    capex_storage_installed = hydrogen_storage_capacity_kg*hydrogen_storage_cost_USDprkg
    capex_compressor_installed = compressor_capex_USDprkWe_of_electrolysis*electrolyzer_size_mw*1000
    #capex_hybrid_installed = hybrid_plant.grid.total_installed_cost
    # capex_hybrid_installed = revised_renewable_cost

    # Fixed and variable costs
    fixed_OM = 12.8 #[$/kW-y]
    fixed_cost_electrolysis_total = fixed_OM*electrolyzer_size_mw*1000
    property_tax_insurance = 1.5/100    #[% of Cap/y]
    variable_OM = 1.30  #[$/MWh]

    H2_PTC_duration = 10 # years the tax credit is active
    Ren_PTC_duration = 10 # years the tax credit is active

    electrolysis_total_EI_policy_grid,electrolysis_total_EI_policy_offgrid\
          = LCA_single_scenario_ProFAST.hydrogen_LCA_singlescenario_ProFAST(grid_connection_scenario,atb_year,site_name,[],[],H2_Results,[],solar_size_mw,storage_size_mw,hopp_dict,H2_PTC_duration)

    AEP_renewables = sum(hopp_dict.main_dict['Models']['grid']['ouput_dict']['energy_from_renewables'])#kWh/year
    grid_cost_keys = list(grid_prices_interpolated_USDperkwh.keys())
    grid_price_per_yr = np.array(list(grid_prices_interpolated_USDperkwh.values()))
    if grid_connection_scenario =='grid-only':
        # grid_electricity_useage_kWhpkg=np.mean(H2_Results['Performance Schedules']['Annual Average Efficiency [kWh/kg]'].values) #new 09/05
        ren_electricity_useage_kWhpkg = 0
        ren_frac = 0

        grid_cost_pr_yr_USDprkg = grid_price_per_yr*energy_used_per_year_kWhpryr/h2prod_per_year_kgpryr
        grid_prices_interpolated_USDperkg = dict(zip(grid_cost_keys,grid_cost_pr_yr_USDprkg))
        
    elif grid_connection_scenario == 'hybrid-grid':
        
        energy_from_grid_pr_yr = energy_used_per_year_kWhpryr - AEP_renewables
        # grid_electricity_useage_kWhpkg=np.mean(energy_from_grid_pr_yr/h2prod_per_year_kgpryr)
        ren_electricity_useage_kWhpkg = AEP_renewables/h2prod_per_year_kgpryr
        wind_electricity_useage_kWhpkg = wind_annual_energy_MWh*1000/h2prod_per_year_kgpryr
        ren_frac = AEP_renewables/energy_used_per_year_kWhpryr
        grid_cost_pr_yr_USDprkg = grid_price_per_yr*energy_from_grid_pr_yr/h2prod_per_year_kgpryr
        grid_prices_interpolated_USDperkg = dict(zip(grid_cost_keys,grid_cost_pr_yr_USDprkg))
        # grid_price_per_yr = np.array(list(grid_prices_interpolated_USDperkwh.values()))
        pass
    elif grid_connection_scenario == 'off-grid':
        # grid_electricity_useage_kWhpkg=0
        ren_frac = 1
        ren_electricity_useage_kWhpkg =AEP_renewables/h2prod_per_year_kgpryr
        wind_electricity_useage_kWhpkg = wind_annual_energy_MWh*1000/h2prod_per_year_kgpryr
        grid_prices_interpolated_USDperkg = dict(zip(grid_cost_keys,np.zeros(len(grid_cost_keys))))
    #new changes end here


    # Establish Ren PTC and assign total emission intensity
    start_year = atb_year + 5
    endofincentives_year = start_year + H2_PTC_duration
    Ren_PTC = {}
    y_idx = 0
    electrolysis_total_EI_policy = {}
    for year in range(start_year,endofincentives_year):
        if atb_year < 2035:
            if grid_connection_scenario == 'grid-only':
                Ren_PTC[year] = 0
                electrolysis_total_EI_policy[year] = electrolysis_total_EI_policy_grid[year]
            elif grid_connection_scenario == 'off-grid':
                electrolysis_total_EI_policy[year] = electrolysis_total_EI_policy_offgrid[year]
                if policy_option == 'no-policy':
                    Ren_PTC[year] = 0
                elif policy_option == 'base':
                    if solar_ITC == True:
                        Ren_PTC[year] = policy['Wind PTC'] * wind_electricity_useage_kWhpkg[y_idx]
                    else:
                        Ren_PTC[year] = policy['Wind PTC'] * ren_electricity_useage_kWhpkg[y_idx]#np.sum(energy_to_electrolyzer)/ (H2_Results['hydrogen_annual_output'])
                elif policy_option == 'max':
                    if solar_ITC == True:
                        Ren_PTC[year] = policy['Wind PTC'] * wind_electricity_useage_kWhpkg[y_idx]
                    else:
                        Ren_PTC[year] = policy['Wind PTC'] * ren_electricity_useage_kWhpkg[y_idx]#np.sum(energy_to_electrolyzer)/ (H2_Results['hydrogen_annual_output'])
            elif grid_connection_scenario == 'hybrid-grid':
                electrolysis_total_EI_policy[year] = 0 # Basically this is not used for hybrid-grid
                if policy_option == 'no-policy':
                    Ren_PTC[year] = 0
                elif policy_option == 'base':
                    if solar_ITC == True:
                        Ren_PTC[year] = policy['Wind PTC'] * wind_electricity_useage_kWhpkg[y_idx]
                    else:
                        Ren_PTC[year] = policy['Wind PTC']  * ren_electricity_useage_kWhpkg[y_idx]#energy_from_renewables / (H2_Results['hydrogen_annual_output'])
                    #Ren_PTC = 0.0051  * np.sum(energy_to_electrolyzer)/ (H2_Results['hydrogen_annual_output']) # We will need to fix this by introducing ren_frac multiplier to denominator when HOPP changes to dealing with grid cases are changed
                elif policy_option == 'max':
                    if solar_ITC == True:
                        Ren_PTC[year] = policy['Wind PTC'] * wind_electricity_useage_kWhpkg[y_idx]
                    else:
                        Ren_PTC[year] = policy['Wind PTC'] * ren_electricity_useage_kWhpkg[y_idx]#energy_from_renewables/ (H2_Results['hydrogen_annual_output'])
                    # Ren_PTC = 0.03072 * np.sum(energy_to_electrolyzer)/ (H2_Results['hydrogen_annual_output']) # We will need to fix this by introducing ren_frac multiplier to denominator when HOPP changes to dealing with grid cases are changed
        elif atb_year == 2035:
                Ren_PTC[year]=0
        y_idx +=1
    # add in electrolzyer replacement schedule
    electrolyzer_refurbishment_schedule = np.zeros(plant_life)
    if user_defined_stack_replacement_time:
        refturb_period = round(user_defined_time_between_replacement/(24*365))
        electrolyzer_refurbishment_schedule[refturb_period:plant_life:refturb_period]=stack_replacement_cost
    else:
        refturb_period = int(np.floor(H2_Results['new_H2_Results']['Time Until Replacement [hrs]']/(24*365)))
        #int(np.floor(H2_Results['avg_time_between_replacement']/(24*365)))
        if replace_stacks_individually:
            percent_of_capacity_replaced = H2_Results['Performance Schedules']['Refurbishment Schedule [MW replaced/year]'].values/electrolyzer_size_mw
            electrolyzer_refurbishment_schedule = percent_of_capacity_replaced*stack_replacement_cost
        else:
            electrolyzer_refurbishment_schedule[refturb_period:plant_life:refturb_period]=stack_replacement_cost
    
    # electrolyzer_refurbishment_schedule[refturb_period:plant_life:refturb_period]=stack_replacement_cost

    # total_variable_OM_perkg = variable_OM*elec_avg_consumption_kWhprkg/1000 #09/05
    annual_variable_OM_perkg = (variable_OM/1000)*elec_efficiency_per_yr_kWhprkg #09/05

    if grid_connection_scenario != 'grid-only':
        wind_om_cost_kw =  renewable_plant_cost_info['wind']['o&m_per_kw']
        fixed_cost_wind = wind_om_cost_kw*wind_size_mw*1000
        capex_wind_installed_init = renewable_plant_cost_info['wind']['capex_per_kw'] * wind_size_mw*1000
        wind_cost_adj = [val for val in renewable_plant_cost_info['wind_savings_dollars'].values()]
        wind_revised_cost=np.sum(wind_cost_adj)

        solar_om_cost_kw = renewable_plant_cost_info['pv']['o&m_per_kw']
        fixed_cost_solar = solar_om_cost_kw*solar_size_mw*1000
        capex_solar_installed = renewable_plant_cost_info['pv']['capex_per_kw'] * solar_size_mw*1000
        battery_hrs=renewable_plant_cost_info['battery']['storage_hours']
        battery_capex_per_kw= renewable_plant_cost_info['battery']['capex_per_kwh']*battery_hrs +  renewable_plant_cost_info['battery']['capex_per_kw']
        capex_battery_installed = battery_capex_per_kw * renewable_plant_cost_info['battery']['size_mw']*1000
        fixed_cost_battery = renewable_plant_cost_info['battery']['o&m_percent'] * capex_battery_installed

    else:
        capex_wind_installed_init=0
        wind_revised_cost = 0
        wind_om_cost_kW = 0
        fixed_cost_solar=0
        capex_solar_installed=0
        capex_battery_installed=0
        fixed_cost_battery=0

    capex_wind_installed=capex_wind_installed_init+wind_revised_cost


    #Calculate H2 and combined PTC
    cambium_year = atb_year + 5
    endofincentives_year = cambium_year + H2_PTC_duration
    H2_PTC = {}
    y_idx = 0
    for year in range(cambium_year,endofincentives_year):
        
        if atb_year < 2035:
        
            if grid_connection_scenario == 'grid-only' or grid_connection_scenario == 'off-grid':

                if policy_option == 'no-policy':
                    ITC = 0
                    H2_PTC[year] = 0 # $/kg H2
                    Ren_PTC[year] = 0 # $/kWh

                elif policy_option == 'max':

                    ITC = 0.5

                    if electrolysis_total_EI_policy[year] <= 0.45: # kg CO2e/kg H2
                        H2_PTC[year] = 3 # $/kg H2
                    elif electrolysis_total_EI_policy[year] > 0.45 and electrolysis_total_EI_policy[year] <= 1.5: # kg CO2e/kg H2
                        H2_PTC[year] = 1 # $/kg H2
                    elif electrolysis_total_EI_policy[year] > 1.5 and electrolysis_total_EI_policy[year] <= 2.5: # kg CO2e/kg H2
                        H2_PTC[year] = 0.75 # $/kg H2
                    elif electrolysis_total_EI_policy[year] > 2.5 and electrolysis_total_EI_policy[year] <= 4: # kg CO2e/kg H2
                        H2_PTC[year] = 0.6 # $/kg H2
                    elif electrolysis_total_EI_policy[year] > 4:
                        H2_PTC[year] = 0

                elif policy_option == 'base':

                    ITC = 0.06

                    if electrolysis_total_EI_policy[year] <= 0.45: # kg CO2e/kg H2
                        H2_PTC[year] = 0.6 # $/kg H2
                    elif electrolysis_total_EI_policy[year] > 0.45 and electrolysis_total_EI_policy[year] <= 1.5: # kg CO2e/kg H2
                        H2_PTC[year] = 0.2 # $/kg H2
                    elif electrolysis_total_EI_policy[year] > 1.5 and electrolysis_total_EI_policy[year] <= 2.5: # kg CO2e/kg H2
                        H2_PTC[year] = 0.15 # $/kg H2
                    elif electrolysis_total_EI_policy[year] > 2.5 and electrolysis_total_EI_policy[year] <= 4: # kg CO2e/kg H2
                        H2_PTC[year] = 0.12 # $/kg H2
                    elif electrolysis_total_EI_policy[year] > 4:
                        H2_PTC[year] = 0


            if grid_connection_scenario == 'hybrid-grid':

                if policy_option == 'no-policy':
                    H2_PTC_grid = 0
                    H2_PTC_offgrid = 0
                    ITC = 0.0

                elif policy_option == 'max':

                    if electrolysis_total_EI_policy_grid[year] <= 0.45: # kg CO2e/kg H2
                        H2_PTC_grid = 3 # $/kg H2
                    elif electrolysis_total_EI_policy_grid[year] > 0.45 and electrolysis_total_EI_policy_grid[year] <= 1.5: # kg CO2e/kg H2
                        H2_PTC_grid = 1 # $/kg H2
                    elif electrolysis_total_EI_policy_grid[year] > 1.5 and electrolysis_total_EI_policy_grid[year] <= 2.5: # kg CO2e/kg H2
                        H2_PTC_grid = 0.75 # $/kg H2
                    elif electrolysis_total_EI_policy_grid[year] > 2.5 and electrolysis_total_EI_policy_grid[year] <= 4: # kg CO2e/kg H2
                        H2_PTC_grid = 0.6 # $/kg H2
                    elif electrolysis_total_EI_policy_grid[year] > 4:
                        H2_PTC_grid = 0

                    if electrolysis_total_EI_policy_offgrid[year] <= 0.45: # kg CO2e/kg H2
                        H2_PTC_offgrid = 3 # $/kg H2
                    elif electrolysis_total_EI_policy_offgrid[year] > 0.45 and electrolysis_total_EI_policy_offgrid[year] <= 1.5: # kg CO2e/kg H2
                        H2_PTC_offgrid = 1 # $/kg H2
                    elif electrolysis_total_EI_policy_offgrid[year] > 1.5 and electrolysis_total_EI_policy_offgrid[year] <= 2.5: # kg CO2e/kg H2
                        H2_PTC_offgrid = 0.75 # $/kg H2
                    elif electrolysis_total_EI_policy_offgrid[year] > 2.5 and electrolysis_total_EI_policy_offgrid[year] <= 4: # kg CO2e/kg H2
                        H2_PTC_offgrid = 0.6 # $/kg H2
                    elif electrolysis_total_EI_policy_offgrid[year] > 4:
                        H2_PTC_offgrid = 0

                    ITC = 0.5

                elif policy_option == 'base':

                    if electrolysis_total_EI_policy_grid[year] <= 0.45: # kg CO2e/kg H2
                        H2_PTC_grid = 0.6 # $/kg H2
                    elif electrolysis_total_EI_policy_grid[year] > 0.45 and electrolysis_total_EI_policy_grid[year] <= 1.5: # kg CO2e/kg H2
                        H2_PTC_grid = 0.2 # $/kg H2
                    elif electrolysis_total_EI_policy_grid[year] > 1.5 and electrolysis_total_EI_policy_grid[year] <= 2.5: # kg CO2e/kg H2
                        H2_PTC_grid = 0.15 # $/kg H2
                    elif electrolysis_total_EI_policy_grid[year] > 2.5 and electrolysis_total_EI_policy_grid[year] <= 4: # kg CO2e/kg H2
                        H2_PTC_grid = 0.12 # $/kg H2
                    elif electrolysis_total_EI_policy_grid[year] > 4:
                        H2_PTC_grid = 0


                    if electrolysis_total_EI_policy_offgrid[year] <= 0.45: # kg CO2e/kg H2
                        H2_PTC_offgrid = 0.6 # $/kg H2
                    elif electrolysis_total_EI_policy_offgrid[year] > 0.45 and electrolysis_total_EI_policy_offgrid[year] <= 1.5: # kg CO2e/kg H2
                        H2_PTC_offgrid = 0.2 # $/kg H2
                    elif electrolysis_total_EI_policy_offgrid[year] > 1.5 and electrolysis_total_EI_policy_offgrid[year] <= 2.5: # kg CO2e/kg H2
                        H2_PTC_offgrid = 0.15 # $/kg H2
                    elif electrolysis_total_EI_policy_offgrid[year] > 2.5 and electrolysis_total_EI_policy_offgrid[year] <= 4: # kg CO2e/kg H2
                        H2_PTC_offgrid = 0.12 # $/kg H2
                    elif electrolysis_total_EI_policy_offgrid[year] > 4:
                        H2_PTC_offgrid = 0
                    
                    ITC = 0.06

                #H2_PTC =  ren_frac * H2_PTC_offgrid + (elec_cf - ren_frac) * H2_PTC_grid
                H2_PTC[year] =  ren_frac[y_idx] * H2_PTC_offgrid + (1 - ren_frac[y_idx]) * H2_PTC_grid
            #combined_PTC[year] = H2_PTC[year]+Ren_PTC[year]
        elif atb_year == 2035:
            H2_PTC[year]=0
        y_idx +=1
    # Reassign PTC values to zero for atb year 2035
    if atb_year == 2035: # need to clarify with Matt when exactly the H2 PTC would end
        #H2_PTC = 0
        #Ren_PTC = 0
        ITC = 0.0
    # if grid_price_scenario == 'retail-flat':
    #     elec_price_perkWh = mwh_to_kwh*elec_price # convert $/MWh to $/kWh
    # Set up ProFAST

    financial_assumptions = pd.read_csv('H2_Analysis/financial_inputs.csv',index_col=None,header=0)
    financial_assumptions.set_index(["Parameter"], inplace = True)

    fraction_capex_vre = (capex_wind_installed + capex_solar_installed + capex_battery_installed)\
                        /(capex_wind_installed + capex_solar_installed + capex_battery_installed + capex_electrolyzer_overnight + capex_desal + capex_compressor_installed + capex_storage_installed)

    fraction_debt_financing_vre = 1/(1+1/financial_assumptions.loc['debt equity ratio of initial financing','Wind/Solar/Bat'])
    fraction_debt_financing_h2 = 1/(1+1/financial_assumptions.loc['debt equity ratio of initial financing','Hydrogen/Steel/Ammonia'])
    fraction_equity_financing_vre = 1 - fraction_debt_financing_vre
    fraction_equity_financing_h2 = 1 - fraction_debt_financing_h2

    real_roe_vre = financial_assumptions.loc['leverage after tax nominal discount rate','Wind/Solar/Bat']
    real_roe_h2 = financial_assumptions.loc['leverage after tax nominal discount rate','Hydrogen/Steel/Ammonia']

    real_interest_vre = financial_assumptions.loc['debt interest rate','Wind/Solar/Bat']
    real_interest_h2 = financial_assumptions.loc['debt interest rate','Hydrogen/Steel/Ammonia']

    real_roe_combined = (fraction_equity_financing_vre*fraction_capex_vre*real_roe_vre + fraction_equity_financing_h2*(1-fraction_capex_vre)*real_roe_h2)\
                        /(fraction_equity_financing_vre*fraction_capex_vre + fraction_equity_financing_h2*(1-fraction_capex_vre))

    real_interest_combined = (real_interest_vre*fraction_capex_vre*fraction_debt_financing_vre + real_interest_h2*(1-fraction_capex_vre)*fraction_debt_financing_h2)\
                             /(fraction_capex_vre*fraction_debt_financing_vre + (1-fraction_capex_vre)*fraction_debt_financing_h2)

    debt_equity_ratio_combined = financial_assumptions.loc['debt equity ratio of initial financing','Wind/Solar/Bat']*fraction_capex_vre + financial_assumptions.loc['debt equity ratio of initial financing','Hydrogen/Steel/Ammonia']*(1-fraction_capex_vre)

    total_income_tax_rate_combined = financial_assumptions.loc['total income tax rate','Wind/Solar/Bat']*fraction_capex_vre + financial_assumptions.loc['total income tax rate','Hydrogen/Steel/Ammonia']*(1-fraction_capex_vre)
    capitalgains_tax_rate_combined = financial_assumptions.loc['capital gains tax rate','Wind/Solar/Bat']*fraction_capex_vre + financial_assumptions.loc['capital gains tax rate','Hydrogen/Steel/Ammonia']*(1-fraction_capex_vre)
    gen_inflation = 0.00

    nominal_roe_combined = (real_roe_combined+1)*(1+gen_inflation)-1
    nominal_interest_combined = (real_interest_combined+1)*(1+gen_inflation)-1
   # total_income_tax_rate = 

    pf = ProFAST.ProFAST()

    # Fill these in - can have most of them as 0 also
    
    install_years = 3
    analysis_start = atb_year + 5 - install_years
    #09/05 (5 lines below)
    operation_start = analysis_start + install_years
    operational_years = np.arange(operation_start,operation_start+plant_life,1)
    year_keys = ['{}'.format(y) for y in operational_years]
    cf_per_year_vals = H2_Results['Performance Schedules']['Capacity Factor [-]'].values #new 09/05
    elec_cf_per_year_PF = dict(zip(year_keys,cf_per_year_vals))
    life_average_cf = H2_Results['Performance Schedules']['Capacity Factor [-]'].mean()
    total_variable_OM_perkg = dict(zip(year_keys,annual_variable_OM_perkg))
    # elec_cf = np.mean(cf_per_year_vals)

    pf.set_params('commodity',{"name":'Hydrogen',"unit":"kg","initial price":100,"escalation":gen_inflation})
    pf.set_params('capacity',electrolysis_plant_capacity_kgperday) #units/day
    pf.set_params('maintenance',{"value":0,"escalation":gen_inflation})
    pf.set_params('analysis start year',analysis_start)
    pf.set_params('operating life',plant_life)
    pf.set_params('installation months',install_years*12)
    pf.set_params('installation cost',{"value":0,"depr type":"Straight line","depr period":4,"depreciable":False})
    pf.set_params('non depr assets',land_cost)
    pf.set_params('end of proj sale non depr assets',land_cost*(1+gen_inflation)**plant_life)
    pf.set_params('demand rampup',0)
    # pf.set_params('long term utilization',elec_cf)
    pf.set_params('long term utilization',elec_cf_per_year_PF) #09/05
    pf.set_params('credit card fees',0)
    pf.set_params('sales tax',0)
    pf.set_params('license and permit',{'value':00,'escalation':gen_inflation})
    pf.set_params('rent',{'value':0,'escalation':gen_inflation})
    pf.set_params('property tax and insurance',property_tax_insurance)
    pf.set_params('admin expense',0)
    pf.set_params('total income tax rate',total_income_tax_rate_combined)
    pf.set_params('capital gains tax rate',capitalgains_tax_rate_combined)
    pf.set_params('sell undepreciated cap',True)
    pf.set_params('tax losses monetized',True)
    pf.set_params('general inflation rate',gen_inflation)
    pf.set_params('leverage after tax nominal discount rate',nominal_roe_combined)
    pf.set_params('debt equity ratio of initial financing',debt_equity_ratio_combined)
    pf.set_params('debt type','Revolving debt')
    pf.set_params('debt interest rate',nominal_interest_combined)
    pf.set_params('cash onhand',1)
    if solar_ITC == True:
        pf.set_params('one time cap inct',{'value':ITC*(capex_storage_installed+capex_battery_installed+capex_solar_installed),'depr type':'MACRS','depr period':7,'depreciable':True})
    else:
        pf.set_params('one time cap inct',{'value':ITC*(capex_storage_installed+capex_battery_installed),'depr type':'MACRS','depr period':7,'depreciable':True})
    #pf.set_params('one time cap inct',{'value':ITC*capex_solar_installed,'depr type':'MACRS','depr period':7,'depreciable':True})
    #pf.set_params('one time cap inct',{'value':ITC*capex_battery_installed,'depr type':'MACRS','depr period':7,'depreciable':True})

    #----------------------------------- Add capital items to ProFAST ----------------
    #pf.add_capital_item(name="Electrolysis system",cost=capex_electrolyzer_overnight,depr_type="MACRS",depr_period=5,refurb=[0])
    pf.add_capital_item(name="Electrolysis system",cost=capex_electrolyzer_overnight,depr_type="MACRS",depr_period=7,refurb=list(electrolyzer_refurbishment_schedule))
    pf.add_capital_item(name="Compression",cost=capex_compressor_installed,depr_type="MACRS",depr_period=7,refurb=[0])
    pf.add_capital_item(name="Hydrogen Storage",cost=capex_storage_installed,depr_type="MACRS",depr_period=7,refurb=[0])
    pf.add_capital_item(name ="Desalination",cost = capex_desal,depr_type="MACRS",depr_period=7,refurb=[0])

    if grid_connection_scenario == 'grid-only':
        pf.add_capital_item(name = "Wind Plant",cost = 0,depr_type = "MACRS",depr_period = 7,refurb = [0])
        pf.add_capital_item(name = "Solar Plant",cost = 0,depr_type = "MACRS",depr_period = 7,refurb = [0])
        pf.add_capital_item(name = "Battery Storage",cost = 0,depr_type = "MACRS",depr_period = 7,refurb = [0])
    else:
        pf.add_capital_item(name = "Wind Plant",cost = capex_wind_installed,depr_type = "MACRS",depr_period = 7,refurb = [0])
        pf.add_capital_item(name = "Solar Plant",cost = capex_solar_installed,depr_type = "MACRS",depr_period = 7,refurb = [0])
        pf.add_capital_item(name = "Battery Storage",cost = capex_battery_installed,depr_type = "MACRS",depr_period = 7,refurb = [0])
        # pf.add_capital_item(name = "Renewable Plant",cost = capex_hybrid_installed,depr_type = "MACRS",depr_period = 5,refurb = [0])
    #replacement_capex = np.sum(electrolyzer_total_installed_capex*electrolyzer_refurbishment_schedule)
    #NOTE TOTAL CAPEX DOES NOT REFLECT STACK REPLACEMENT COSTS AND HOW THEY CHANGE OVER TIME!
    total_capex = capex_electrolyzer_overnight+capex_compressor_installed+capex_storage_installed+capex_desal+capex_wind_installed+capex_solar_installed + capex_battery_installed #+ replacement_capex#capex_hybrid_installed
    # total_capex = capex_electrolyzer_overnight+capex_compressor_installed+capex_storage_installed+capex_desal+ capex_hybrid_installed
    # capex_fraction = {'Electrolyzer':capex_electrolyzer_overnight/total_capex,
    #                   'Compression':capex_compressor_installed/total_capex,
    #                   'Hydrogen Storage':capex_storage_installed/total_capex,
    #                   'Desalination':capex_desal/total_capex,
    #                   'Wind Plant':capex_wind_installed/total_capex,
    #                   'Solar Plant':capex_solar_installed/total_capex,
    #                   'Battery Storage':capex_battery_installed/total_capex,
    #                   #'Stack Replacement': replacement_capex/total_capex
    #                   #'Renewable Plant':capex_hybrid_installed/total_capex
    #                   }

    #-------------------------------------- Add fixed costs--------------------------------
    pf.add_fixed_cost(name="Electrolyzer Fixed O&M Cost",usage=1.0,unit='$/year',cost=fixed_cost_electrolysis_total,escalation=gen_inflation)
    pf.add_fixed_cost(name="Desalination Fixed O&M Cost",usage=1.0,unit='$/year',cost=opex_desal,escalation=gen_inflation)
    #pf.add_fixed_cost(name="Renewable Plant Fixed O&M Cost",usage=1.0,unit='$/year',cost=fixed_cost_renewables,escalation=gen_inflation)

    if grid_connection_scenario == 'grid-only':
        # pf.add_fixed_cost(name="Renewable Plant Fixed O&M Cost",usage=1.0,unit='$/year',cost=0,escalation=gen_inflation)
        pf.add_fixed_cost(name="Wind Plant Fixed O&M Cost",usage=1.0,unit='$/year',cost=0,escalation=gen_inflation)
        pf.add_fixed_cost(name="Solar Plant Fixed O&M Cost",usage=1.0,unit='$/year',cost=0,escalation=gen_inflation)
        pf.add_fixed_cost(name="Battery Storage Fixed O&M Cost",usage=1.0,unit='$/year',cost=0,escalation=gen_inflation)
    else:
        # pf.add_fixed_cost(name="Renewable Plant Fixed O&M Cost",usage=1.0,unit='$/year',cost=fixed_cost_renewables,escalation=gen_inflation)
        pf.add_fixed_cost(name="Wind Plant Fixed O&M Cost",usage=1.0,unit='$/year',cost=fixed_cost_wind,escalation=gen_inflation)
        pf.add_fixed_cost(name="Solar Plant Fixed O&M Cost",usage=1.0,unit='$/year',cost=fixed_cost_solar,escalation=gen_inflation)
        pf.add_fixed_cost(name="Battery Storage Fixed O&M Cost",usage=1.0,unit='$/year',cost=fixed_cost_battery,escalation=gen_inflation)

    
    #---------------------- Add feedstocks, note the various cost options-------------------
    #pf.add_feedstock(name='Electricity',usage=elec_avg_consumption_kWhprkg,unit='kWh',cost=lcoe/100,escalation=gen_inflation)
    pf.add_feedstock(name='Water',usage=water_consumption_avg_galH2O_prkgH2,unit='gallon-water',cost=water_cost,escalation=gen_inflation)
    pf.add_feedstock(name='Var O&M',usage=1.0,unit='$/kg',cost=total_variable_OM_perkg,escalation=gen_inflation)

    #TODO: include changes in energy required/year for grid usage
    # pf.add_feedstock(name='Grid Electricity Cost',usage=grid_electricity_useage_kWhpkg,unit='$/kWh',cost=grid_prices_interpolated_USDperkwh,escalation=gen_inflation)
    pf.add_feedstock(name='Grid Electricity Cost',usage=1.0,unit='$/kg',cost=grid_prices_interpolated_USDperkg,escalation=gen_inflation)
    
    # pf.add_feedstock(name='Grid Electricity Cost',usage=min_grid_usage,unit='$/kWh',cost=grid_prices_interpolated_USDperkwh,escalation=gen_inflation)
    #---------------------- Add various tax credit incentives -------------------
    pf.add_incentive(name ='Renewable PTC credit', value=Ren_PTC, decay = 0, sunset_years = Ren_PTC_duration, tax_credit = True)
    pf.add_incentive(name ='Hydrogen PTC credit', value=H2_PTC, decay = 0, sunset_years = H2_PTC_duration, tax_credit = True)

    sol = pf.solve_price()

    summary = pf.summary_vals

    price_breakdown = pf.get_cost_breakdown()

    # Calculate contribution of equipment to breakeven price
    total_price_capex = price_breakdown.loc[price_breakdown['Name']=='Electrolysis system','NPV'].tolist()[0]\
                      + price_breakdown.loc[price_breakdown['Name']=='Compression','NPV'].tolist()[0]\
                      + price_breakdown.loc[price_breakdown['Name']=='Hydrogen Storage','NPV'].tolist()[0]\
                      + price_breakdown.loc[price_breakdown['Name']=='Desalination','NPV'].tolist()[0]\
                      + price_breakdown.loc[price_breakdown['Name']=='Wind Plant','NPV'].tolist()[0]\
                      + price_breakdown.loc[price_breakdown['Name']=='Solar Plant','NPV'].tolist()[0]\
                      + price_breakdown.loc[price_breakdown['Name']=='Battery Storage','NPV'].tolist()[0]\

    capex_fraction = {'Electrolyzer':price_breakdown.loc[price_breakdown['Name']=='Electrolysis system','NPV'].tolist()[0]/total_price_capex,
                  'Compression':price_breakdown.loc[price_breakdown['Name']=='Compression','NPV'].tolist()[0]/total_price_capex,
                  'Hydrogen Storage':price_breakdown.loc[price_breakdown['Name']=='Hydrogen Storage','NPV'].tolist()[0]/total_price_capex,
                  'Desalination':price_breakdown.loc[price_breakdown['Name']=='Desalination','NPV'].tolist()[0]/total_price_capex,
                  'Wind Plant':price_breakdown.loc[price_breakdown['Name']=='Wind Plant','NPV'].tolist()[0]/total_price_capex,
                  'Solar Plant':price_breakdown.loc[price_breakdown['Name']=='Solar Plant','NPV'].tolist()[0]/total_price_capex,
                  'Battery Storage':price_breakdown.loc[price_breakdown['Name']=='Battery Storage','NPV'].tolist()[0]/total_price_capex}

    # Calculate financial expense associated with equipment
    cap_expense = price_breakdown.loc[price_breakdown['Name']=='Repayment of debt','NPV'].tolist()[0]\
        + price_breakdown.loc[price_breakdown['Name']=='Interest expense','NPV'].tolist()[0]\
        + price_breakdown.loc[price_breakdown['Name']=='Dividends paid','NPV'].tolist()[0]\
        - price_breakdown.loc[price_breakdown['Name']=='Inflow of debt','NPV'].tolist()[0]\
        - price_breakdown.loc[price_breakdown['Name']=='Inflow of equity','NPV'].tolist()[0]

    # Calculate remaining financial expenses
    remaining_financial = price_breakdown.loc[price_breakdown['Name']=='Non-depreciable assets','NPV'].tolist()[0]\
        + price_breakdown.loc[price_breakdown['Name']=='Cash on hand reserve','NPV'].tolist()[0]\
        + price_breakdown.loc[price_breakdown['Name']=='Property insurance','NPV'].tolist()[0]\
        - price_breakdown.loc[price_breakdown['Name']=='Sale of non-depreciable assets','NPV'].tolist()[0]\
        - price_breakdown.loc[price_breakdown['Name']=='Cash on hand recovery','NPV'].tolist()[0]

    # Calculate LCOH breakdown and assign capital expense to equipment costs
    price_breakdown_electrolyzer = price_breakdown.loc[price_breakdown['Name']=='Electrolysis system','NPV'].tolist()[0] + cap_expense*capex_fraction['Electrolyzer']
    price_breakdown_compression = price_breakdown.loc[price_breakdown['Name']=='Compression','NPV'].tolist()[0] + cap_expense*capex_fraction['Compression']
    price_breakdown_storage = price_breakdown.loc[price_breakdown['Name']=='Hydrogen Storage','NPV'].tolist()[0]+cap_expense*capex_fraction['Hydrogen Storage']
    price_breakdown_desalination = price_breakdown.loc[price_breakdown['Name']=='Desalination','NPV'].tolist()[0] + cap_expense*capex_fraction['Desalination']
    # price_breakdown_renewables = price_breakdown.loc[price_breakdown['Name']=='Renewable Plant','NPV'].tolist()[0] + cap_expense*capex_fraction['Renewable Plant']
    price_breakdown_wind = price_breakdown.loc[price_breakdown['Name']=='Wind Plant','NPV'].tolist()[0] + cap_expense*capex_fraction['Wind Plant']
    price_breakdown_solar = price_breakdown.loc[price_breakdown['Name']=='Solar Plant','NPV'].tolist()[0] + cap_expense*capex_fraction['Solar Plant']
    price_breakdown_battery = price_breakdown.loc[price_breakdown['Name']=='Battery Storage','NPV'].tolist()[0] + cap_expense*capex_fraction['Battery Storage']


    price_breakdown_electrolysis_FOM = price_breakdown.loc[price_breakdown['Name']=='Electrolyzer Fixed O&M Cost','NPV'].tolist()[0]
    price_breakdown_electrolysis_VOM = price_breakdown.loc[price_breakdown['Name']=='Var O&M','NPV'].tolist()[0]
    price_breakdown_desalination_FOM = price_breakdown.loc[price_breakdown['Name']=='Desalination Fixed O&M Cost','NPV'].tolist()[0]
    # price_breakdown_renewables_FOM = price_breakdown.loc[price_breakdown['Name']=='Renewable Plant Fixed O&M Cost','NPV'].tolist()[0]
    price_breakdown_wind_FOM = price_breakdown.loc[price_breakdown['Name']=='Wind Plant Fixed O&M Cost','NPV'].tolist()[0]
    price_breakdown_solar_FOM = price_breakdown.loc[price_breakdown['Name']=='Solar Plant Fixed O&M Cost','NPV'].tolist()[0]
    price_breakdown_battery_FOM = price_breakdown.loc[price_breakdown['Name']=='Battery Storage Fixed O&M Cost','NPV'].tolist()[0]
    price_breakdown_taxes = price_breakdown.loc[price_breakdown['Name']=='Income taxes payable','NPV'].tolist()[0]\
        - price_breakdown.loc[price_breakdown['Name'] == 'Monetized tax losses','NPV'].tolist()[0]\

    if gen_inflation > 0:
        price_breakdown_taxes = price_breakdown_taxes + price_breakdown.loc[price_breakdown['Name']=='Capital gains taxes payable','NPV'].tolist()[0]

    price_breakdown_water = price_breakdown.loc[price_breakdown['Name']=='Water','NPV'].tolist()[0]

    price_breakdown_grid_elec_price = price_breakdown.loc[price_breakdown['Name']=='Grid Electricity Cost','NPV'].tolist()[0]

    # price_breakdown_financial = price_breakdown.loc[price_breakdown['Name']=='Non-depreciable assets','NPV'].tolist()[0]\
    #     + price_breakdown.loc[price_breakdown['Name']=='Cash on hand reserve','NPV'].tolist()[0]\
    #     + price_breakdown.loc[price_breakdown['Name']=='Property insurance','NPV'].tolist()[0]\
    #     + price_breakdown.loc[price_breakdown['Name']=='Repayment of debt','NPV'].tolist()[0]\
    #     + price_breakdown.loc[price_breakdown['Name']=='Interest expense','NPV'].tolist()[0]\
    #     + price_breakdown.loc[price_breakdown['Name']=='Dividends paid','NPV'].tolist()[0]\
    #     - price_breakdown.loc[price_breakdown['Name']=='Sale of non-depreciable assets','NPV'].tolist()[0]\
    #     - price_breakdown.loc[price_breakdown['Name']=='Cash on hand recovery','NPV'].tolist()[0]\
    #     - price_breakdown.loc[price_breakdown['Name']=='Inflow of debt','NPV'].tolist()[0]\
    #     - price_breakdown.loc[price_breakdown['Name']=='Inflow of equity','NPV'].tolist()[0]
    price_breakdown_renewables=price_breakdown_wind + price_breakdown_solar +price_breakdown_battery
    price_breakdown_renewables_FOM = price_breakdown_wind_FOM + price_breakdown_solar_FOM + price_breakdown_battery_FOM
    lcoh_check = price_breakdown_electrolyzer+price_breakdown_compression+price_breakdown_storage+price_breakdown_electrolysis_FOM\
        + price_breakdown_desalination+price_breakdown_desalination_FOM+ price_breakdown_electrolysis_VOM\
            +price_breakdown_renewables+price_breakdown_renewables_FOM+price_breakdown_taxes+price_breakdown_water+price_breakdown_grid_elec_price+remaining_financial\
                #+ price_breakdown_stack_replacement

    lcoh_breakdown = {'LCOH: Compression & storage ($/kg)':price_breakdown_storage+price_breakdown_compression,\
                      'LCOH: Electrolyzer CAPEX ($/kg)':price_breakdown_electrolyzer,'LCOH: Desalination CAPEX ($/kg)':price_breakdown_desalination,\
                      'LCOH: Electrolyzer FOM ($/kg)':price_breakdown_electrolysis_FOM,'LCOH: Electrolyzer VOM ($/kg)':price_breakdown_electrolysis_VOM,\
                      'LCOH: Desalination FOM ($/kg)':price_breakdown_desalination_FOM,\
                      'LCOH: Wind Plant ($/kg)':price_breakdown_wind,'LCOH: Wind Plant FOM ($/kg)':price_breakdown_wind_FOM,\
                      'LCOH: Solar Plant ($/kg)':price_breakdown_solar,'LCOH: Solar Plant FOM ($/kg)':price_breakdown_solar_FOM,\
                      'LCOH: Battery Storage ($/kg)':price_breakdown_battery,'LCOH: Battery Storage FOM ($/kg)':price_breakdown_battery_FOM,\
                      #'LCOH: Renewable plant ($/kg)':price_breakdown_renewables,'LCOH: Renewable FOM ($/kg)':price_breakdown_renewables_FOM,
                      'LCOH: Taxes ($/kg)':price_breakdown_taxes,\
                      'LCOH: Water consumption ($/kg)':price_breakdown_water,'LCOH: Grid electricity ($/kg)':price_breakdown_grid_elec_price,\
                      'LCOH: Finances ($/kg)':remaining_financial,'LCOH: total ($/kg)':lcoh_check,'LCOH Profast:':sol['price']}
    print('{} {} {} has LCOH = ${} /kg'.format(site_name,policy_option,grid_connection_scenario,sol['price']))
    price_breakdown = price_breakdown.drop(columns=['index','Amount'])
   
    # return(sol,summary,price_breakdown,lcoh_breakdown,capex_electrolyzer_overnight/electrolyzer_size_mw/1000,elec_cf,ren_frac,electrolysis_total_EI_policy_grid,electrolysis_total_EI_policy_offgrid,H2_PTC,Ren_PTC,total_capex)
    return(sol,summary,price_breakdown,lcoh_breakdown,capex_electrolyzer_overnight/electrolyzer_size_mw/1000,life_average_cf,ren_frac,electrolysis_total_EI_policy_grid,electrolysis_total_EI_policy_offgrid,H2_PTC,Ren_PTC,total_capex)


import os
import sys
sys.path.append('')
from dotenv import load_dotenv
import pandas as pd
import json
from hopp.simulation.technologies.sites import SiteInfo
from hopp.simulation.technologies.sites import flatirons_site as sample_site
from hopp.utilities.keys import set_developer_nrel_gov_key
from greenheart.to_organize.H2_Analysis import simple_dispatch
from greenheart.to_organize.H2_Analysis.simple_cash_annuals import simple_cash_annuals
import numpy as np
import matplotlib.pyplot as plt
import warnings
from pathlib import Path
warnings.filterwarnings("ignore")

from greenheart.to_organize import hopp_tools
from greenheart.to_organize import inputs_py
import copy 
from greenheart.to_organize import plot_results

"""
Perform a LCOE analysis for a few locations across the U.S. to demonstrate analysis across locations  using HOPP
A few notes:
# 1. physical interaction effects ignored currently 
# 2. shared BOS to be re-implemented
"""

#Set API key
load_dotenv()
NREL_API_KEY = os.getenv("NREL_API_KEY")
set_developer_nrel_gov_key('NREL_API_KEY')  # Set this key manually here if you are not setting it using the .env

#Step 1: User Inputs for scenario
resource_year = 2013
atb_years = [
            2020,
            2030,
            2050
            ]
policy = {
    'option 1': {'Wind ITC': 0, 'Wind PTC': 0},
    'option 2': {'Wind ITC': 26, 'Wind PTC': 0},
    'option 3': {'Wind ITC': 6, 'Wind PTC': 0},
    'option 4': {'Wind ITC': 30, 'Wind PTC': 0},
    'option 5': {'Wind ITC': 50, 'Wind PTC': 0},
}

sample_site['year'] = resource_year
useful_life = 30
critical_load_factor = 1
run_reopt_flag = False
custom_powercurve = True    #A flag that is applicable when using PySam WindPower (not FLORIS)
storage_used = True
battery_can_grid_charge = False
grid_connected_hopp = False
floris = True

# Technology sizing
interconnection_size_mw = 1000
wind_size_mw = 1000
solar_size_mw = 500
storage_size_mw = 100
storage_size_mwh = 400 # 100 MW, 4 hr

scenario_choice = 'Example HOPP buildout'

site_selection = [
                'Site 1',
                'Site 2',
                'Site 3',
                'Site 4'
                ]
scenario = dict()
kw_continuous = (wind_size_mw + solar_size_mw) * 1000
load = [kw_continuous for x in
        range(0, 8760)]  # * (sin(x) + pi) Set desired/required load profile for plant

#Site lat and lon will be set by data loaded from Orbit runs

# Financial inputs
discount_rate = 0.07
debt_equity_split = 60

# These inputs are not used in this analysis (no solar or storage)
solar_cost_kw = 640  # TODO: take from input file for different years (fine if just looking at current year)
storage_cost_kw = 1500
storage_cost_kwh = 380

# Flags (TODO: update documentation)
forced_sizes = True # no REopt

# Enable Ability to purchase/sell electricity to/from grid. Price Defined in $/kWh
# sell_price = 0.01
# buy_price = 0.01
sell_price = False
buy_price = False

# Set paths for results, floris and orbit
parent_path = os.path.abspath('')
results_dir = parent_path + '/examples/hybrids/results/'
floris_dir = parent_path + '/floris_input_files/'

print('Parent path = ', parent_path)

# Site specific turbine information
path = ('examples/hybrids/location_based_costs.xlsx')
xl = pd.ExcelFile(path)

# outputs
save_outputs_dict = inputs_py.establish_save_output_dict()
save_all_runs = list()

# which plots to show
plot_power_production = True
plot_battery = True
plot_grid = True
plot_wind = True
print_results = True


for i in policy:
    # set policy values
    scenario, policy_option = hopp_tools.set_policy_values(scenario, policy, i)
    print('Wind PTC: ', scenario['Wind PTC'])

    for atb_year in atb_years:
        for site_location in site_selection:
            scenario_df = xl.parse()
            scenario_df.set_index(["Parameter"], inplace = True)
            
            site_df = scenario_df[site_location]

            turbine_model = str(site_df['Turbine Rating'])+'MW'

            # set turbine values
            scenario, nTurbs, floris_config = hopp_tools.set_turbine_model(turbine_model, scenario, parent_path,floris_dir, floris)

            scenario['Useful Life'] = useful_life

            # financials
            scenario = hopp_tools.set_financial_info(scenario, debt_equity_split, discount_rate)

            # site info
            site_df, sample_site = hopp_tools.set_site_info(site_df, sample_site)
            site_name = site_df['State']
            site = SiteInfo(sample_site, hub_height=scenario['Tower Height'])

            #Assign scenario cost details (TODO: hard-coded from spreadsheet - not a standardized way)
            if atb_year == 2020:
                total_capex = site_df['2020 CapEx']
                wind_om_cost_kw = site_df['2020 OpEx ($/kw-yr)']
            if atb_year == 2030:
                total_capex = site_df['2030 CapEx']
                wind_om_cost_kw = site_df['2030 OpEx ($/kw-yr)']
            if atb_year == 2050:
                total_capex = site_df['2050 CapEx']
                wind_om_cost_kw = site_df['2050 OpEx ($/kw-yr)']

            capex_multiplier = site_df['CapEx Multiplier']
            wind_cost_kw = copy.deepcopy(total_capex) * capex_multiplier

            #Plot Wind Data to ensure offshore data is sound
            wind_data = site.wind_resource._data['data']
            wind_speed = [W[2] for W in wind_data]
            plot_results.plot_wind_results(wind_data, site_name, site_df['Representative coordinates'], results_dir, plot_wind)

            
            # Run HOPP
            combined_hybrid_power_production_hopp, energy_shortfall_hopp, combined_hybrid_curtailment_hopp, hybrid_plant, wind_size_mw, solar_size_mw, lcoe = \
                hopp_tools.run_HOPP(scenario,
                                    site,
                                    sample_site,
                                    forced_sizes,
                                    solar_size_mw,
                                    wind_size_mw,
                                    storage_size_mw,
                                    storage_size_mwh,
                                    wind_cost_kw, 
                                    solar_cost_kw, 
                                    storage_cost_kw, 
                                    storage_cost_kwh,
                                    kw_continuous, 
                                    load,
                                    interconnection_size_mw,
                                    wind_om_cost_kw,
                                    nTurbs,
                                    floris_config,
                                    floris)

            generation_summary_df = pd.DataFrame({'Generation profile (kW)': hybrid_plant.grid.generation_profile[0:8760] })
            generation_summary_df.to_csv(os.path.join(results_dir, 'Generation Summary_{}_{}_{}_{}.csv'.format(site_name,atb_year,turbine_model,scenario['Powercurve File'])))

            #Step 4: Plot HOPP Results
            plot_results.plot_HOPP(combined_hybrid_power_production_hopp,
                                    energy_shortfall_hopp,
                                    combined_hybrid_curtailment_hopp,
                                    load,
                                    results_dir,
                                    site_name,
                                    atb_year,
                                    turbine_model,
                                    hybrid_plant,
                                    plot_power_production)

            #Step 5: Run Simple Dispatch Model
            combined_pv_wind_storage_power_production_hopp, battery_SOC, battery_used, excess_energy = \
                hopp_tools.run_battery(energy_shortfall_hopp, combined_hybrid_curtailment_hopp, combined_hybrid_power_production_hopp)
            
            plot_results.plot_battery_results(combined_hybrid_curtailment_hopp, 
                        energy_shortfall_hopp,
                        combined_pv_wind_storage_power_production_hopp,
                        combined_hybrid_power_production_hopp,
                        battery_SOC,
                        battery_used,
                        results_dir,
                        site_name,atb_year,turbine_model,
                        load,
                        plot_battery)

            

            # grid information (on and off-grid systems)
            cost_to_buy_from_grid, profit_from_selling_to_grid, energy_to_electrolyzer = hopp_tools.grid(combined_pv_wind_storage_power_production_hopp,
                                                                                    sell_price,
                                                                                    excess_energy,
                                                                                    buy_price,
                                                                                    kw_continuous,
                                                                                    plot_grid)

            # calculate financials simple (no H2)
            total_elec_production = np.sum(combined_pv_wind_storage_power_production_hopp) 
            cf = total_elec_production / (interconnection_size_mw * 1000 * 8760)
            # plt.plot(energy_to_electrolyzer)
            # plt.show()

            print(scenario)

            if print_results:
                # ------------------------- #
                #TODO: Tidy up these print statements
                print("Future Scenario: {}".format(site_location))
                print("Wind CapEx Cost per KW: {}".format(total_capex))
                print("PV Cost per KW: {}".format(solar_cost_kw))
                print("Storage Cost per KW: {}".format(storage_cost_kw))
                print("Storage Cost per KWh: {}".format(storage_cost_kwh))
                print("Wind Size built: {}".format(wind_size_mw))
                print("PV Size built: {}".format(solar_size_mw))
                print("Storage Size built: {}".format(storage_size_mw))
                print("Storage Size built: {}".format(storage_size_mwh))
                print("Total Yearly Electrical Output: {}".format(total_elec_production))
                print('Combined Capacity Factor: ', cf)
                print("Levelized cost of Electricity (cents/kWh): {}".format(lcoe))
                print('=========================================================')


print('Done')


# implement a simple dispatch model

import numpy as np

class SimpleDispatch():

    def __init__(self):

        # length of simulation
        self.Nt = 1
        
        # amount of curtailment experienced by plant
        self.curtailment = np.zeros(self.Nt)

        # amount of energy needed from the battery
        self.shortfall = np.zeros(self.Nt)
        
        # size of battery (MWh)
        self.battery_storage = 0

        # Charge rate of the battery (MW)
        self.charge_rate = 0
        self.discharge_rate = 0
        

    def run(self):

        # storage module
        battery_storage = self.battery_storage  # MWh
        charge_rate = self.charge_rate  # MW
        discharge_rate = self.discharge_rate #MW

        battery_SOC = np.zeros(self.Nt)
        battery_used = np.zeros(self.Nt)
        excess_energy = np.zeros(self.Nt)
        
        for i in range(self.Nt):
            # should you charge
            if self.curtailment[i] > 0:
                if i == 0:
                    battery_SOC[i] = np.min([self.curtailment[i], charge_rate])
                    amount_charged = battery_SOC[i]
                    excess_energy[i] = self.curtailment[i] - amount_charged
                else:
                    if battery_SOC[i-1] < battery_storage:
                        add_gen = np.min([self.curtailment[i], charge_rate])
                        battery_SOC[i] = np.min([battery_SOC[i-1] + add_gen, battery_storage])
                        amount_charged = battery_SOC[i] - battery_SOC[i-1]
                        excess_energy[i] = self.curtailment[i] - amount_charged
                    else:
                        battery_SOC[i] = battery_SOC[i - 1]
                        excess_energy[i] = self.curtailment[i]

            # should you discharge
            else:
                if i > 0:
                    if battery_SOC[i-1] > 0:
                        
                        battery_used[i] = np.min([self.shortfall[i], battery_SOC[i-1],discharge_rate])
                        battery_SOC[i] = battery_SOC[i-1] - battery_used[i]

        # print('==============================================')
        # print('Battery Generation: ', np.sum(battery_used))
        # print('Amount of energy going to the grid: ', np.sum(excess_energy))
        # print('==============================================')
        return battery_used, excess_energy, battery_SOC






from greenheart.to_organize.H2_Analysis.simple_cash_annuals import simple_cash_annuals

class Compressor():
    def __init__(self, input_dict, output_dict):
        self.input_dict = input_dict
        self.output_dict = output_dict

        # inputs
        self.flow_rate_kg_hr = input_dict['flow_rate_kg_hr']                        #[kg/hr] # per compressor
        self.P_outlet = input_dict['P_outlet']                                      #[bar]
        self.compressor_rating_kWe = input_dict['compressor_rating_kWe']            #[kWe] # per compressor
        self.mean_time_between_failure = input_dict['mean_time_between_failure']    #[days]
        self.total_hydrogen_throughput = input_dict['total_hydrogen_throughput']    #[kg-H2/yr]

        try:
            number_of_compressors = input_dict["number_of_compressors"]
        except:
            print("Assuming 3 compressors")
            number_of_compressors = 3
        try:
            plant_life = input_dict["plant_life"]
        except:
            print("Assuming 30 year plant life")
            plant_life = 30
        try:
            useful_life = input_dict["useful_life"]
        except:
            print("Assuming 15 year useful life")
            useful_life = 15

        # assumptions
        self.comp_efficiency = 0.50
        self.num_compressors = number_of_compressors # was 3
        self.plant_life = plant_life # [years], was 30
        self.useful_life = useful_life # [years], was 15

    def compressor_power(self):
        """ Compression from 20 bar to 250 bar (pressure vessel storage)
            or compression from 20 bar to 100 bar (underground pipe storage)
            https://www.energy.gov/sites/default/files/2014/03/f9/nexant_h2a.pdf
        TODO: Add CoolProp to be able to calculate all power for different compressions"""
        Z = 1.03198     # mean compressibility factor
        R = 4.1240      # [kJ/kg K] hydrogen gas constant
        k = 1.667       # ratio of specific heats
        T = 25+273.15   # [C] suction and interstage gas temperature
        P_inlet = 20    # [bar] from electrolyzer

        if self.P_outlet == 250 or self.P_outlet == 100:    #[bar]
            comp_energy_per_kg = Z * R * T * (1/self.comp_efficiency) * (k/(k-1)) * ((self.P_outlet/P_inlet)**((k-1)/k)-1) / 3600     # [kWh/kg -per compressor]
            compressor_power = self.num_compressors * self.flow_rate_kg_hr * comp_energy_per_kg #[kW] - cumulative across all compressors
        else:
            raise(ValueError("Error. P_outlet must be 100 or 250 bar."))
        self.output_dict['comp_energy_per_kg'] = comp_energy_per_kg
        self.output_dict['compressor_power'] = compressor_power
        return comp_energy_per_kg, compressor_power

    def compressor_costs(self):
        """Minimum 2 compressors required due to unreliability"""
        F_install = 1.2     # installation factor (<250 kg/hr)
        F_install_250 = 2.0     # installation factor (>250 kg/hr)
        F_indir = 1.27      # direct and indirect capital cost factor

        C_cap = 19207*self.compressor_rating_kWe**(0.6089) # purchased equipment capital cost

        if self.flow_rate_kg_hr < 250:
            compressor_capex = C_cap * F_install * F_indir * self.num_compressors #[USD]
        else:
            compressor_capex = C_cap * F_install_250 * F_indir * self.num_compressors #[USD]
        self.output_dict['compressor_capex'] = compressor_capex

        #Compressor opex
        insurance = 0.01    #percent of total capital investment
        property_taxes = 0.015  #percent of total capital investment
        license_permits = 0.01  #percent of total capital investment
        op_and_maint = 0.04     #percent of total installed capital
        labor = 4.2977 * self.flow_rate_kg_hr**0.2551
        overhead = 0.5 * labor

        cost_factors = insurance + property_taxes + license_permits
        compressor_opex = (cost_factors * C_cap) + (op_and_maint * compressor_capex) + labor + overhead

        # """"TODO: Add mean_time_between_failure [days]: max 365
        #     total_hydrogen_throughput: annual amount of hydrogen compressed [kg/yr]
        #     This report gives station costs as a function of MTBF but not broken down to single compressor level
        #     https://www.nrel.gov/docs/fy14osti/58564.pdf"""
        # if self.mean_time_between_failure <= 50:       #[days]
        #     maintenance_cost = 0.71     #[USD/kg H2]
        #     compressor_opex = self.num_compressors * maintenance_cost * self.total_hydrogen_throughput  #[USD/yr]
        # elif 50 < self.mean_time_between_failure <= 100:
        #     maintenance_cost = 0.71 + ((self.mean_time_between_failure - 50)*((0.36 - 0.71)/(100-50)))     #[USD/kg H2]
        #     compressor_opex = self.num_compressors * maintenance_cost * self.total_hydrogen_throughput  #[USD/yr]
        # elif 100 < self.mean_time_between_failure <= 200:
        #     maintenance_cost = 0.36 + ((self.mean_time_between_failure - 100)*((0.19 - 0.36)/(200-100)))     #[USD/kg H2]
        #     compressor_opex = self.num_compressors * maintenance_cost * self.total_hydrogen_throughput  #[USD/yr]
        # elif 200 < self.mean_time_between_failure <= 365:
        #     maintenance_cost = 0.11 + ((self.mean_time_between_failure - 200)*((0.11 - 0.19)/(365-200)))     #[USD/kg H2]
        #     compressor_opex = self.num_compressors * maintenance_cost * self.total_hydrogen_throughput  #[USD/yr]
        # else:
        #     print("Error. mean_time_between_failure <= 365 days.")
        self.output_dict['compressor_opex'] = compressor_opex

        """Assumed useful life = payment period for capital expenditure.
           compressor amortization interest = 3%"""

        compressor_annuals = simple_cash_annuals(self.plant_life, self.useful_life,\
            self.output_dict['compressor_capex'],self.output_dict['compressor_opex'], 0.03)

        self.output_dict['compressor_annuals'] = compressor_annuals
        return compressor_capex, compressor_opex, compressor_annuals

if __name__ =="__main__":

    in_dict = dict()
    in_dict['flow_rate_kg_hr'] = 126
    in_dict['P_outlet'] = 250
    in_dict['compressor_rating_kWe'] = 802
    in_dict['mean_time_between_failure'] = 200
    in_dict['total_hydrogen_throughput'] = 5000000
    out_dict = dict()

    test = Compressor(in_dict, out_dict)
    test.compressor_power()
    test.compressor_costs()
    print("compressor_power (kW): ", out_dict['compressor_power'])
    print("Compressor capex [USD]: ", out_dict['compressor_capex'])
    print("Compressor opex [USD/yr]: ", out_dict['compressor_opex'])
    print(out_dict['compressor_annuals'])


## Low-Temperature PEM Electrolyzer Model
"""
Python model of H2 PEM low-temp electrolyzer.

Quick Hydrogen Physics:

1 kg H2 <-> 11.1 N-m3 <-> 33.3 kWh (LHV) <-> 39.4 kWh (HHV)

High mass energy density (1 kg H2= 3,77 l gasoline)
Low volumetric density (1 Nm³ H2= 0,34 l gasoline

Hydrogen production from water electrolysis (~5 kWh/Nm³ H2)

Power:1 MW electrolyser <-> 200 Nm³/h  H2 <-> ±18 kg/h H2
Energy:+/-55 kWh of electricity --> 1 kg H2 <-> 11.1 Nm³ <-> ±10 liters
demineralized water

Power production from a hydrogen PEM fuel cell from hydrogen (+/-50%
efficiency):
Energy: 1 kg H2 --> 16 kWh
"""
import math
import numpy as np
import sys
import pandas as pd
from matplotlib import pyplot as plt

np.set_printoptions(threshold=sys.maxsize)


class PEM_electrolyzer_LT:
    """
    Create an instance of a low-temperature PEM Electrolyzer System. Each
    stack in the electrolyzer system in this model is rated at 1 MW_DC.

    Parameters
    _____________
    np_array P_input_external_kW
        1-D array of time-series external power supply

    string voltage_type
        Nature of voltage supplied to electrolyzer from the external power
        supply ['variable' or 'constant]

    float power_supply_rating_MW
        Rated power of external power supply

    Returns
    _____________

    """

    def __init__(self, input_dict, output_dict):
        self.input_dict = input_dict
        self.output_dict = output_dict

        # array of input power signal
        self.input_dict['P_input_external_kW'] = input_dict['P_input_external_kW']
        self.electrolyzer_system_size_MW = input_dict['electrolyzer_system_size_MW']

        # self.input_dict['voltage_type'] = 'variable'  # not yet implemented
        self.input_dict['voltage_type'] = 'constant'
        self.stack_input_voltage_DC = 250

        # Assumptions:
        self.min_V_cell = 1.62  # Only used in variable voltage scenario
        self.p_s_h2_bar = 31  # H2 outlet pressure
        self.stack_input_current_lower_bound = 500
        self.stack_rating_kW = 1000  # 1 MW
        self.cell_active_area = 1250
        self.N_cells = 130

        # Constants:
        self.moles_per_g_h2 = 0.49606
        self.V_TN = 1.48  # Thermo-neutral Voltage (Volts)
        self.F = 96485  # Faraday's Constant (C/mol)
        self.R = 8.314  # Ideal Gas Constant (J/mol/K)

        self.external_power_supply()

    def external_power_supply(self):
        """
        External power source (grid or REG) which will need to be stepped
        down and converted to DC power for the electrolyzer.

        Please note, for a wind farm as the electrolyzer's power source,
        the model assumes variable power supplied to the stack at fixed
        voltage (fixed voltage, variable power and current)

        TODO: extend model to accept variable voltage, current, and power
        This will replicate direct DC-coupled PV system operating at MPP
        """
        power_converter_efficiency = 0.95
        if self.input_dict['voltage_type'] == 'constant':

            self.input_dict['P_input_external_kW'] = \
                np.where(self.input_dict['P_input_external_kW'] >
                         (self.electrolyzer_system_size_MW * 1000),
                         (self.electrolyzer_system_size_MW * 1000),
                         self.input_dict['P_input_external_kW'])

            self.output_dict['curtailed_P_kW'] = \
                np.where(self.input_dict['P_input_external_kW'] >
                         (self.electrolyzer_system_size_MW * 1000),
                         (self.input_dict['P_input_external_kW'] -
                          (self.electrolyzer_system_size_MW * 1000)), 0)

            self.output_dict['current_input_external_Amps'] = \
                (self.input_dict['P_input_external_kW'] * 1000 *
                 power_converter_efficiency) / (self.stack_input_voltage_DC *
                                                self.system_design())

            self.output_dict['stack_current_density_A_cm2'] = \
                self.output_dict['current_input_external_Amps'] / self.cell_active_area

            self.output_dict['current_input_external_Amps'] = \
                np.where(self.output_dict['current_input_external_Amps'] <
                         self.stack_input_current_lower_bound, 0,
                         self.output_dict['current_input_external_Amps'])

        else:
            pass  # TODO: extend model to variable voltage and current source

    def system_design(self):
        """
        For now, system design is solely a function of max. external power
        supply; i.e., a rated power supply of 50 MW means that the electrolyzer
        system developed by this model is also rated at 50 MW

        TODO: Extend model to include this capability.
        Assume that a PEM electrolyzer behaves as a purely resistive load
        in a circuit, and design the configuration of the entire electrolyzer
        system - which may consist of multiple stacks connected together in
        series, parallel, or a combination of both.
        """
        h2_production_multiplier = (self.electrolyzer_system_size_MW * 1000) / \
                                   self.stack_rating_kW
        self.output_dict['electrolyzer_system_size_MW'] = self.electrolyzer_system_size_MW
        return h2_production_multiplier

    def cell_design(self):
        """
        Creates an I-V (polarization) curve of each cell in a stack.

        Please note that this method is currently not used in the model. It
        will be used once the electrolyzer model is expanded to variable
        voltage supply as well as implementation of the self.system_design()
        method

        Motivation:

        The most common representation of the electrolyzer performance is the
        polarization curve that represents the relation between the current density
        and the voltage (V):
        Source: https://www.sciencedirect.com/science/article/pii/S0959652620312312

        V = N_c(E_cell + V_Act,c + V_Act,a + iR_cell)

        where N_c is the number of electrolyzer cells,E_cell is the open circuit
        voltage VAct,and V_Act,c are the anode and cathode activation over-potentials,
        i is the current density and iRcell is the electrolyzer cell resistance
        (ohmic losses).

        Use this to make a V vs. A (Amperes/cm2) graph which starts at 1.23V because
        thermodynamic reaction of water formation/splitting dictates that standard
        electrode potential has a ∆G of 237 kJ/mol (where: ∆H = ∆G + T∆S)
        """

        # Cell level inputs:
        N_cells = 130
        electrode_surface_area_cm2 = self.cell_active_area / N_cells
        cell_rating_watts = (self.stack_rating_kW * 1000) / N_cells

        # V_cell_max = 3.0    #Volts
        # V_cell_I_density_max = 2.50     #mA/cm2
        E_rev = 1.23  # (in Volts) Reversible potential at 25degC
        E_th = 1.48  # (in Volts) Thermoneutral potential at 25degC
        T_C = 80  # Celsius
        T_K = T_C + 273.15  # in Kelvins
        # E_cell == Open Circuit Voltage
        E_cell = 1.5184 - (1.5421 * (10 ** (-3)) * T_K) + \
                 (9.523 * (10 ** (-5)) * T_K * math.log(T_K)) + \
                 (9.84 * (10 ** (-8)) * (T_K ** 2))
        # V_act = V_act_c + V_Act_a
        R = 8.314  # Ideal Gas Constant (J/mol/K)
        i = self.output_dict['stack_current_density_A_cm2']
        F = 96485  # Faraday's Constant (C/mol)

        # Following coefficient values obtained from Yigit and Selamet (2016) -
        # https://www.sciencedirect.com/science/article/pii/S0360319916318341?via%3Dihub
        a_a = 2  # Anode charge transfer coefficient
        a_c = 0.5  # Cathode charge transfer coefficient
        i_o_a = 2 * (10 ** (-7))
        i_o_c = 2 * (10 ** (-3))
        V_act = (((R * T_K) / (a_a * F)) * np.arcsinh(i / (2 * i_o_a))) + (
                ((R * T_K) / (a_c * F)) * np.arcsinh(i / (2 * i_o_c)))
        lambda_water_content = ((-2.89556 + (0.016 * T_K)) + 1.625) / 0.1875
        delta = 0.0003  # membrane thickness (cm) - assuming a 3-µm thick membrane
        sigma = ((0.005139 * lambda_water_content) - 0.00326) * math.exp(
            1268 * ((1 / 303) - (1 / T_K)))  # Material thickness # material conductivity
        R_cell = (delta / sigma)
        V_cell = E_cell + V_act + (i * R_cell)
        V_cell = np.where(V_cell < E_rev, E_rev, V_cell)
        V_stack = N_cells * V_cell  # Stack operational voltage

    def dynamic_operation(self):
        """
        Model the electrolyzer's realistic response/operation under variable RE

        TODO: add this capability to the model
        """
        # When electrolyzer is already at or near its optimal operation
        # temperature (~80degC)
        warm_startup_time_secs = 30
        cold_startup_time_secs = 5 * 60  # 5 minutes

    def water_electrolysis_efficiency(self):
        """
        https://www.sciencedirect.com/science/article/pii/S2589299119300035#b0500

        According to the first law of thermodynamics energy is conserved.
        Thus, the conversion efficiency calculated from the yields of
        converted electrical energy into chemical energy. Typically,
        water electrolysis efficiency is calculated by the higher heating
        value (HHV) of hydrogen. Since the electrolysis process water is
        supplied to the cell in liquid phase efficiency can be calculated by:

        n_T = V_TN / V_cell

        where, V_TN is the thermo-neutral voltage (min. required V to
        electrolyze water)

        Parameters
        ______________

        Returns
        ______________

        """

        n_T = self.V_TN / (self.stack_input_voltage_DC / self.N_cells)
        return n_T

    def faradaic_efficiency(self):
        """`
        Text background from:
        [https://www.researchgate.net/publication/344260178_Faraday%27s_
        Efficiency_Modeling_of_a_Proton_Exchange_Membrane_Electrolyzer_
        Based_on_Experimental_Data]

        In electrolyzers, Faraday’s efficiency is a relevant parameter to
        assess the amount of hydrogen generated according to the input
        energy and energy efficiency. Faraday’s efficiency expresses the
        faradaic losses due to the gas crossover current. The thickness
        of the membrane and operating conditions (i.e., temperature, gas
        pressure) may affect the Faraday’s efficiency.

        Equation for n_F obtained from:
        https://www.sciencedirect.com/science/article/pii/S0360319917347237#bib27

        Parameters
        ______________
        float f_1
            Coefficient - value at operating temperature of 80degC (mA2/cm4)

        float f_2
            Coefficient - value at operating temp of 80 degC (unitless)

        np_array current_input_external_Amps
            1-D array of current supplied to electrolyzer stack from external
            power source


        Returns
        ______________

        float n_F
            Faradaic efficiency (unitless)

        """
        f_1 = 250  # Coefficient (mA2/cm4)
        f_2 = 0.996  # Coefficient (unitless)
        I_cell = self.output_dict['current_input_external_Amps'] * 1000

        # Faraday efficiency
        n_F = (((I_cell / self.cell_active_area) ** 2) /
               (f_1 + ((I_cell / self.cell_active_area) ** 2))) * f_2

        return n_F

    def compression_efficiency(self):
        """
        In industrial contexts, the remaining hydrogen should be stored at
        certain storage pressures that vary depending on the intended
        application. In the case of subsequent compression, pressure-volume
        work, Wc, must be performed. The additional pressure-volume work can
        be related to the heating value of storable hydrogen. Then, the total
        efficiency reduces by the following factor:
        https://www.mdpi.com/1996-1073/13/3/612/htm

        Due to reasons of material properties and operating costs, large
        amounts of gaseous hydrogen are usually not stored at pressures
        exceeding 100 bar in aboveground vessels and 200 bar in underground
        storages
        https://www.sciencedirect.com/science/article/pii/S0360319919310195

        Partial pressure of H2(g) calculated using:
        The hydrogen partial pressure is calculated as a difference between
        the  cathode  pressure, 101,325 Pa, and the water saturation
        pressure
        [Source: Energies2018,11,3273; doi:10.3390/en11123273]

        """
        n_limC = 0.825  # Limited efficiency of gas compressors (unitless)
        H_LHV = 241  # Lower heating value of H2 (kJ/mol)
        K = 1.4  # Average heat capacity ratio (unitless)
        C_c = 2.75  # Compression factor (ratio of pressure after and before compression)
        n_F = self.faradaic_efficiency()
        j = self.output_dict['stack_current_density_A_cm2']
        n_x = ((1 - n_F) * j) * self.cell_active_area
        n_h2 = j * self.cell_active_area
        Z = 1  # [Assumption] Average compressibility factor (unitless)
        T_in_C = 80  # Assuming electrolyzer operates at 80degC
        T_in = 273.15 + T_in_C  # (Kelvins) Assuming electrolyzer operates at 80degC
        W_1_C = (K / (K - 1)) * ((n_h2 - n_x) / self.F) * self.R * T_in * Z * \
                ((C_c ** ((K - 1) / K)) - 1)  # Single stage compression

        # Calculate partial pressure of H2 at the cathode:
        A = 8.07131
        B = 1730.63
        C = 233.426
        p_h2o_sat = 10 ** (A - (B / (C + T_in_C)))  # Pa
        p_cat = 101325  # Cathode pressure (Pa)
        p_h2_cat = p_cat - p_h2o_sat
        p_s_h2_Pa = self.p_s_h2_bar * 1e5

        s_C = math.log((p_s_h2_Pa / p_h2_cat), 10) / math.log(C_c, 10)
        W_C = round(s_C) * W_1_C  # Pressure-Volume work - energy reqd. for compression
        net_energy_carrier = n_h2 - n_x  # C/s
        net_energy_carrier = np.where((n_h2 - n_x) == 0, 1, net_energy_carrier)
        n_C = 1 - ((W_C / (((net_energy_carrier) / self.F) * H_LHV * 1000)) * (1 / n_limC))
        n_C = np.where((n_h2 - n_x) == 0, 0, n_C)
        return n_C

    def total_efficiency(self):
        """
        Aside from efficiencies accounted for in this model
        (water_electrolysis_efficiency, faradaic_efficiency, and
        compression_efficiency) all process steps such as gas drying above
        2 bar or water pumping can be assumed as negligible. Ultimately, the
        total efficiency or system efficiency of a PEM electrolysis system is:

        n_T = n_p_h2 * n_F_h2 * n_c_h2
        https://www.mdpi.com/1996-1073/13/3/612/htm
        """
        n_p_h2 = self.water_electrolysis_efficiency()
        n_F_h2 = self.faradaic_efficiency()
        n_c_h2 = self.compression_efficiency()

        n_T = n_p_h2 * n_F_h2 * n_c_h2
        self.output_dict['total_efficiency'] = n_T
        return n_T

    def h2_production_rate(self):
        """
        H2 production rate calculated using Faraday's Law of Electrolysis
        (https://www.sciencedirect.com/science/article/pii/S0360319917347237#bib27)

        Parameters
        _____________

        float f_1
            Coefficient - value at operating temperature of 80degC (mA2/cm4)

        float f_2
            Coefficient - value at operating temp of 80 degC (unitless)

        np_array
            1-D array of current supplied to electrolyzer stack from external
            power source


        Returns
        _____________

        """
        # Single stack calculations:
        n_Tot = self.total_efficiency()
        h2_production_rate = n_Tot * ((self.N_cells *
                                       self.output_dict['current_input_external_Amps']) /
                                      (2 * self.F))  # mol/s
        h2_production_rate_g_s = h2_production_rate / self.moles_per_g_h2
        h2_produced_kg_hr = h2_production_rate_g_s * 3.6 * 72.55/55.5 ## TEMPORARY CORRECTION APPLIED FOR PEM EFFICIENCY to reach expected 55.5kwh/kg value

        self.output_dict['stack_h2_produced_kg_hr'] = h2_produced_kg_hr

        # Total electrolyzer system calculations:
        h2_produced_kg_hr_system = self.system_design() * h2_produced_kg_hr
        # h2_produced_kg_hr_system = h2_produced_kg_hr
        self.output_dict['h2_produced_kg_hr_system'] = h2_produced_kg_hr_system

        return h2_produced_kg_hr_system

    def degradation(self):
        """
        TODO
        Add a time component to the model - for degradation ->
        https://www.hydrogen.energy.gov/pdfs/progress17/ii_b_1_peters_2017.pdf
        """
        pass

    def water_supply(self):
        """
        Calculate water supply rate based system efficiency and H2 production
        rate
        TODO: Add this capability to the model
        """
        max_water_feed_mass_flow_rate_kg_hr = 411  # kg per hour
        water_used_kg_hr_system = self.h2_production_rate() * 10
        self.output_dict['water_used_kg_hr'] = water_used_kg_hr_system
        self.output_dict['water_used_kg_annual'] = np.sum(water_used_kg_hr_system)

    def h2_storage(self):
        """
        Model to estimate Ideal Isorthermal H2 compression at 70degC
        https://www.sciencedirect.com/science/article/pii/S036031991733954X

        The amount of hydrogen gas stored under pressure can be estimated
        using the van der Waals equation

        p = [(nRT)/(V-nb)] - [a * ((n^2) / (V^2))]

        where p is pressure of the hydrogen gas (Pa), n the amount of
        substance (mol), T the temperature (K), and V the volume of storage
        (m3). The constants a and b are called the van der Waals coefficients,
        which for hydrogen are 2.45 × 10−2 Pa m6mol−2 and 26.61 × 10−6 ,
        respectively.
        """

        pass


if __name__=="__main__":
    # Example on how to use this model:
    in_dict = dict()
    in_dict['electrolyzer_system_size_MW'] = 15
    out_dict = dict()

    electricity_profile = pd.read_csv('sample_wind_electricity_profile.csv')
    in_dict['P_input_external_kW'] = electricity_profile.iloc[:, 1].to_numpy()

    el = PEM_electrolyzer_LT(in_dict, out_dict)
    el.h2_production_rate()
    print("Hourly H2 production by stack (kg/hr): ", out_dict['stack_h2_produced_kg_hr'][0:50])
    print("Hourly H2 production by system (kg/hr): ", out_dict['h2_produced_kg_hr_system'][0:50])
    fig, axs = plt.subplots(2, 2)
    fig.suptitle('PEM H2 Electrolysis Results for ' +
                str(out_dict['electrolyzer_system_size_MW']) + ' MW System')

    axs[0, 0].plot(out_dict['stack_h2_produced_kg_hr'])
    axs[0, 0].set_title('Hourly H2 production by stack')
    axs[0, 0].set_ylabel('kg_h2 / hr')
    axs[0, 0].set_xlabel('Hour')

    axs[0, 1].plot(out_dict['h2_produced_kg_hr_system'])
    axs[0, 1].set_title('Hourly H2 production by system')
    axs[0, 1].set_ylabel('kg_h2 / hr')
    axs[0, 1].set_xlabel('Hour')

    axs[1, 0].plot(in_dict['P_input_external_kW'])
    axs[1, 0].set_title('Hourly Energy Supplied by Wind Farm (kWh)')
    axs[1, 0].set_ylabel('kWh')
    axs[1, 0].set_xlabel('Hour')

    total_efficiency = out_dict['total_efficiency']
    system_h2_eff = (1 / total_efficiency) * 33.3
    system_h2_eff = np.where(total_efficiency == 0, 0, system_h2_eff)

    axs[1, 1].plot(system_h2_eff)
    axs[1, 1].set_title('Total Stack Energy Usage per mass net H2')
    axs[1, 1].set_ylabel('kWh_e/kg_h2')
    axs[1, 1].set_xlabel('Hour')

    plt.show()
    print("Annual H2 production (kg): ", np.sum(out_dict['h2_produced_kg_hr_system']))
    print("Annual energy production (kWh): ", np.sum(in_dict['P_input_external_kW']))
    print("H2 generated (kg) per kWH of energy generated by wind farm: ",
          np.sum(out_dict['h2_produced_kg_hr_system']) / np.sum(in_dict['P_input_external_kW']))



import matplotlib.pyplot as plt

def plot_battery(titletext, df_mean, df_ci, y, ylim, colors, xticks_major, xticks_minor,
                       xlabels_major, xlabels_minor,  save_location):

    fig2, ax2 = plt.subplots(figsize=(15, 10))
    ax3 = ax2.twinx()
    for i in y:
        if i % 24 == 0:
            ax3.plot(y[i: i + 24], df_mean.battery_soc_pct[i: i + 24], marker="o", label="$Battery SOC$", c=colors[1])
            ax3.fill_between(
                y[i: i + 24],
                (df_mean.battery_soc_pct - df_ci.battery_soc_pct)[i: i + 24],
                (df_mean.battery_soc_pct + df_ci.battery_soc_pct)[i: i + 24],
                alpha=0.3, color=colors[1], label="$Battery SOC$ 95% CI"
            )

            ax2.plot(y[i: i + 24], df_mean.wind_power_to_battery[i: i + 24], marker="o", label="$Wind to Battery$", c=colors[0])
            ax2.fill_between(
                y[i: i + 24],
                (df_mean.wind_power_to_battery - df_ci.wind_power_to_battery)[i: i + 24],
                (df_mean.wind_power_to_battery + df_ci.wind_power_to_battery)[i: i + 24],
                alpha=0.3, color=colors[0], label="$Wind to Battery$ 95% CI"
            )
            ax2.plot(y[i: i + 24], df_mean.pv_power_to_battery[i: i + 24], marker="o", label="$Solar to Battery$", c='red')
            ax2.fill_between(
                y[i: i + 24],
                (df_mean.pv_power_to_battery - df_ci.pv_power_to_battery)[i: i + 24],
                (df_mean.pv_power_to_battery + df_ci.pv_power_to_battery)[i: i + 24],
                alpha=0.3, color='red', label="$Solar to Battery$ 95% CI"
            )
            ax2.plot(y[i: i + 24], df_mean.storage_power_to_load[i: i + 24], marker="o", label="$Battery to Load$", c='green')
            ax2.fill_between(
                y[i: i + 24],
                (df_mean.storage_power_to_load - df_ci.storage_power_to_load)[i: i + 24],
                (df_mean.storage_power_to_load + df_ci.storage_power_to_load)[i: i + 24],
                alpha=0.3, color='green', label="$Battery to Load$ 95% CI"
            )
            ax2.plot(y[i: i + 24], df_mean.storage_power_to_load[i: i + 24], marker="o", label="$Battery to Grid$", c='cyan')
            ax2.fill_between(
                y[i: i + 24],
                (df_mean.storage_power_to_load - df_ci.storage_power_to_load)[i: i + 24],
                (df_mean.storage_power_to_load + df_ci.storage_power_to_load)[i: i + 24],
                alpha=0.3, color='cyan', label="$Battery to Grid$ 95% CI"
            )

    ax3.set_ylabel("Battery SOC")
    ax3.set_ylim(0, 1)
    ax2.set_ylabel("Power (kW)")
    ax2.set_xlim(0, 24)
    ax2.set_xticks(xticks_major)
    for t in ax2.get_xticklabels():
        t.set_y(-0.05)
    ax2.set_xticks(xticks_minor, minor=True)
    ax2.set_xticklabels(xlabels_major, ha="right")
    ax2.set_xticklabels(xlabels_minor, minor=True)
    ax2.set_xlabel("Hour of day")
    ax2.set_title(titletext)

    handles, labels = ax2.get_legend_handles_labels()
    labels_set = ["$Wind to Battery$", "$Wind to Battery$ 95% CI",
                   "$Solar to Battery$", "$Solar to Battery$ 95% CI",
                   "$Battery to Load$", "$Battery to Load$ 95% CI",
                   "$Battery to Grid$", "$Battery to Grid$ 95% CI"]

    ax2.grid(alpha=0.7)
    ax2.grid(alpha=0.2, which="minor")

    ix_filter = [labels.index(el) for el in labels_set]
    handles = [handles[ix] for ix in ix_filter]
    labels = [labels[ix] for ix in ix_filter]
    ax2.legend(handles, labels, ncol=3, loc="lower left")

    handles2, labels2 = ax3.get_legend_handles_labels()
    labels_set2 = ["$Battery SOC$", "$Battery SOC$ 95% CI",]
    ix_filter2 = [labels2.index(el) for el in labels_set2]
    handles2 = [handles2[ix] for ix in ix_filter2]
    labels2 = [labels2[ix] for ix in ix_filter2]

    ax2.legend(handles, labels, ncol=3, loc="lower left")

    fig2.tight_layout()
    # fig2.show()
    fig2.savefig(save_location, dpi=240,
                    bbox_to_inches="tight")
    plt.close('all')


from greenheart.simulation.technologies.hydrogen.h2_storage.pressure_vessel.compressed_gas_storage_model_20221021.Compressed_all import PressureVessel
from pytest import approx
import numpy as np

import matplotlib.pyplot as plt

# test that we the results we got when the code was recieved
class TestPressureVessel():


    # pressure_vessel_instance_no_cost = PressureVessel(Energy_cost=0.0)
    # pressure_vessel_instance_no_cost.run()
    pressure_vessel_instance = PressureVessel(Energy_cost=0.07)
    pressure_vessel_instance.run()

    pressure_vessel_instance_no_cost = PressureVessel(Energy_cost=0.0)
    pressure_vessel_instance_no_cost.run()
        
    def test_capacity_max(self):
        assert self.pressure_vessel_instance.capacity_max == 5585222.222222222
    
    def test_t_discharge_hr_max(self):
        assert self.pressure_vessel_instance.t_discharge_hr_max == 25.133499999999998

    def test_a_fit_capex(self):
        # assert self.pressure_vessel_instance.a_fit_capex == 9084.035219940572
        assert self.pressure_vessel_instance.a_fit_capex == approx(0.053925726563169414)
    
    def test_b_fit_capex(self):
        # assert self.pressure_vessel_instance.b_fit_capex == -0.127478041731842
        assert self.pressure_vessel_instance.b_fit_capex == approx(1.6826965840450498)
    
    def test_c_fit_capex(self):
        assert self.pressure_vessel_instance.c_fit_capex == approx(20.297862568544417)

    def test_a_fit_opex(self):
        assert self.pressure_vessel_instance.a_fit_opex == approx(0.05900068374896024)

    def test_b_fit_opex(self):
        assert self.pressure_vessel_instance.b_fit_opex == approx(1.8431485607717895)

    def test_c_fit_opex(self):
        assert self.pressure_vessel_instance.c_fit_opex == approx(17.538017086792006)

    def test_energy_fit(self):
        capacity = 1E6 # 1000 tonnes h2
        _, _, energy_per_kg = self.pressure_vessel_instance.calculate_from_fit(capacity_kg=capacity)
        assert energy_per_kg == approx(2.688696, 1E-5) # kWh/kg

    def test_compare_price_change_capex(self):
        capacity = 1E6 # 1000 tonnes h2
        capex_07, _, _ = self.pressure_vessel_instance.calculate_from_fit(capacity_kg=capacity)
        capex_00, _, _ = self.pressure_vessel_instance_no_cost.calculate_from_fit(capacity_kg=capacity)

        assert capex_00 == capex_07

    def test_compare_price_change_opex(self):
        capacity = 1E6 # 1000 tonnes h2
        _, opex_07, _ = self.pressure_vessel_instance.calculate_from_fit(capacity_kg=capacity)
        _, opex_00, _ = self.pressure_vessel_instance_no_cost.calculate_from_fit(capacity_kg=capacity)

        assert opex_00 < opex_07

    def test_compare_price_change_energy(self):
        capacity = 1E6 # 1000 tonnes h2
        _, _, energy_per_kg_07 = self.pressure_vessel_instance.calculate_from_fit(capacity_kg=capacity)
        _, _, energy_per_kg_00 = self.pressure_vessel_instance_no_cost.calculate_from_fit(capacity_kg=capacity)

        assert energy_per_kg_00 == energy_per_kg_07

    def test_mass_footprint(self):
        """
        extension of gold standard test to new tank footprint outputs
        """
        cap_H2_tank_ref= 179.6322517351785
        capacity_req= 1.2e3
        Ntank_ref= np.ceil(capacity_req/cap_H2_tank_ref)
        Atank_ref= (2*53.7e-2)**2
        footprint_ref= Ntank_ref*Atank_ref
        mass_ref= 1865.0

        assert self.pressure_vessel_instance.get_tanks(capacity_req) == approx(Ntank_ref)
        assert self.pressure_vessel_instance.get_tank_footprint(capacity_req)[0] == approx(Atank_ref, rel= 0.01)
        assert self.pressure_vessel_instance.get_tank_footprint(capacity_req)[1] == approx(footprint_ref, rel= 0.01)

        assert self.pressure_vessel_instance.get_tank_mass(capacity_req)[0] == approx(mass_ref, rel= 0.01)

    def test_output_function(self):
        capacity = self.pressure_vessel_instance.compressed_gas_function.capacity_1[5]
        capex, opex, energy = self.pressure_vessel_instance.calculate_from_fit(capacity)
        tol = 1.0

        assert capex == approx(self.pressure_vessel_instance.compressed_gas_function.cost_kg[5]*capacity, tol)
        assert opex == approx(self.pressure_vessel_instance.compressed_gas_function.Op_c_Costs_kg[5]*capacity, tol)
        assert energy == approx(self.pressure_vessel_instance.compressed_gas_function.total_energy_used_kwh[5]*capacity, tol)

    def test_distributed(self):
        capacity = self.pressure_vessel_instance.compressed_gas_function.capacity_1[5]
        capex, opex, energy_kg = self.pressure_vessel_instance.calculate_from_fit(capacity)
        print(capex)
        mass_tank_empty = self.pressure_vessel_instance.get_tank_mass(capacity)
        area_footprint = self.pressure_vessel_instance.get_tank_footprint(capacity)
        
        capex_dist_05, opex_dist_05, energy_kg_dist_05, area_footprint_site_05, mass_tank_empty_site_05, capacity_site_05= \
                self.pressure_vessel_instance.distributed_storage_vessels(capacity, 5)
        assert capex_dist_05 == approx(6205232868.4722595)
        assert opex_dist_05 == approx(113433768.86938927)
        assert energy_kg_dist_05 == approx(2.6886965443907727)
        assert area_footprint_site_05 == approx(4866.189496204457)
        assert mass_tank_empty_site_05 == approx(7870274.025926539)
        assert capacity_site_05 == approx(capacity/5)

        capex_dist_10, opex_dist_10, energy_kg_dist_10, area_footprint_site_10, mass_tank_empty_site_10, capacity_site_10= \
                self.pressure_vessel_instance.distributed_storage_vessels(capacity, 10)
        assert capex_dist_10 == approx(7430302244.729572)
        assert opex_dist_10 == approx(138351814.3102437)
        assert energy_kg_dist_10 == approx(2.6886965443907727)
        assert area_footprint_site_10 == approx(2433.0947481022286)
        assert mass_tank_empty_site_10 == approx(3935137.0129632694)
        assert capacity_site_10 == approx(capacity/10)

        capex_dist_20, opex_dist_20, energy_kg_dist_20, area_footprint_site_20, mass_tank_empty_site_20, capacity_site_20= \
                self.pressure_vessel_instance.distributed_storage_vessels(capacity, 20)
        assert capex_dist_20 == approx(9370417735.496975)
        assert opex_dist_20 == approx(178586780.2083488)
        assert energy_kg_dist_20 == approx(2.6886965443907727)
        assert area_footprint_site_20 == approx(1216.5473740511143)
        assert mass_tank_empty_site_20 == approx(1967568.5064816347)
        assert capacity_site_20 == approx(capacity/20)

        assert (capex < capex_dist_05) and (capex_dist_05 < capex_dist_10) and (capex_dist_10 < capex_dist_20), "capex should increase w/ number of sites"
        assert (opex < opex_dist_05) and (opex_dist_05 < opex_dist_10) and (opex_dist_10 < opex_dist_20), "opex should increase w/ number of sites"
        assert (energy_kg == approx(energy_kg_dist_05)) and (energy_kg_dist_05 == approx(energy_kg_dist_10)) and (energy_kg_dist_10 == approx(energy_kg_dist_20)), "energy_kg be approx. equal across number of sites"

        # assert False

    # def test_plots(self):
    #     self.pressure_vessel_instance.plot()

class PlotTestPressureVessel():
    def __init__(self) -> None:
        
        self.pressure_vessel_instance = PressureVessel(Energy_cost=0.07)
        self.pressure_vessel_instance.run()

        self.pressure_vessel_instance_no_cost = PressureVessel(Energy_cost=0.0)
        self.pressure_vessel_instance_no_cost.run()

    def plot_size_and_divisions(self):

        # set sweep ranges
        capacity_range = np.arange(1E6, 5E6, step=1E3) # kg
        divisions = np.array([1, 5, 10, 15, 20])

        # initialize outputs
        capex_results = np.zeros((len(divisions), len(capacity_range)))
        opex_results =  np.zeros((len(divisions), len(capacity_range)))

        # run capex and opex for range
        for j, div, in enumerate(divisions):
            for i, capacity in enumerate(capacity_range):
                if div == 1:
                    capex_results[j, i], opex_results[j, i], energy_kg = self.pressure_vessel_instance.calculate_from_fit(capacity)
                else:
                    capex_results[j, i], opex_results[j, i], energy_kg_dist_05, area_footprint_site_05, mass_tank_empty_site_05, capacity_site_05= \
                        self.pressure_vessel_instance.distributed_storage_vessels(capacity, div)

        # plot results
        fig, ax = plt.subplots(2,2, sharex=True)
        for j in np.arange(0, len(divisions)):
            ax[0, 0].plot(capacity_range/1E3, capex_results[j]/capacity_range, label="%i Divisions" %(divisions[j]))
            ax[0, 1].plot(capacity_range/1E3, opex_results[j]/capacity_range, label="%i Divisions" %(divisions[j]))
    
        # ax[0,0].set(ylabel="CAPEX (USD/kg)", ylim=[0, 5000], yticks=[0,1000,2000,3000,4000,5000])
        ax[0,1].set(ylabel="OPEX (USD/kg)", ylim=[0, 100])

        for j in np.arange(0, len(divisions)):
            ax[1, 0].plot(capacity_range/1E3, capex_results[j]/capex_results[0], label="%i Divisions" %(divisions[j]))
            ax[1, 1].plot(capacity_range/1E3, opex_results[j]/opex_results[0], label="%i Divisions" %(divisions[j]))

        ax[1,0].set(ylabel="CAPEX_div/CAPEX_cent", ylim=[0,4])
        ax[1,1].set(ylabel="OPEX_div/OPEX_cent", ylim=[0,4])

        for axi in ax[1]:
            axi.set(xlabel="H$_2$ Capacity (tonnes)")
        ax[1,1].legend()

        plt.tight_layout()

        plt.show()

if __name__ == "__main__":
    # test_set = TestPressureVessel()
    plot_tests = PlotTestPressureVessel()
    plot_tests.plot_size_and_divisions()
    
# 0.0
# 6322420.744236805
# 1331189.5844818645
# 7363353.502353448

# 0.07
# 6322420.744236805
# 1331189.5844818645
# 7363353.502353448

# energy cost for both cases match as per above


# op costs - 0.07
# 442569.45209657634
# 345243.94167843653
# 0
# 93183.27091373052

# op costs - 0.0
# 0.0
# 0.0
# 0
# 0.0

# op c costs
# op_c_costs 0.07
# 880996.6646887433 
#  799322.4503233839 
#  0.03 
#  4262490675.039804 
#  25920

# op_c_costs 0.00
# 0.0 
#  0.0 
#  0.03 
#  4262490675.039804 
#  25920


import pytest
import os
from pathlib import Path

import ORBIT as orbit
from greenheart.simulation.technologies.offshore.floating_platform import install_platform, calc_platform_opex, calc_substructure_mass_and_cost, DesignPhase, InstallPhase

'''Sources:
    - [1] M. Maness, B. Maples and A. Smith, "NREL Offshore Balance-of-System Model," National Renewable Energy Laboratory, 2017. https://www.nrel.gov/docs/fy17osti/66874.pdf
'''

@pytest.fixture
def config():

    offshore_path = Path(__file__).parents[3] / "greenheart" / "simulation" / "technologies" / "offshore"

    return orbit.load_config(os.path.join(offshore_path, "example_floating_project.yaml"))

def test_install_platform(config):
    '''
    Test the code that calculates the platform installation cost
    [1]: equations (91),(113),(98)
    '''
    distance = 24
    mass = 2100
    area = 500

    cost = install_platform(mass, area, distance, install_duration=14, foundation="floating")

    assert pytest.approx(cost) == 7142871
    
def test_calc_substructure_cost(config):
    '''
    Test the code that calculates the CapEx from floating_platform.py
    [1]: equations (81),(83),(84)
    '''
    topmass = 200
    toparea = 1000
    depth = 500
    
    cost,_ = calc_substructure_mass_and_cost(topmass, toparea, depth,fab_cost_rate=14500, design_cost=4500000, sub_cost_rate=3000, line_cost=850000, anchor_cost=120000, anchor_mass=20, line_mass=100000, num_lines=4)
    
    assert pytest.approx(cost) == 23040000
    

def test_calc_substructure_mass(config):
    '''
    Test the code that calculates the CapEx from floating_platform.py
    [1]: equations (81),(83),(84)
    '''
    topmass = 200
    toparea = 1000
    depth = 500
    
    _, mass = calc_substructure_mass_and_cost(topmass, toparea, depth,fab_cost_rate=14500, design_cost=4500000, sub_cost_rate=3000, line_cost=850000, anchor_cost=120000, anchor_mass=20, line_mass=100000, num_lines=4)
    
    assert pytest.approx(mass,.1) == 680.

def test_calc_platform_opex():
    '''
    Test the code that calculates the OpEx from floating_platform.py
    '''
    capex = 28e6
    opex_rate = 0.01
    cost = calc_platform_opex(capex, opex_rate)

    assert pytest.approx(cost) == 28e4


import os
import sys
sys.path.append('')
# from dotenv import load_dotenv
import pandas as pd
# from PEM_H2_LT_electrolyzer_ESGBasicClusters import PEM_electrolyzer_LT as PEMClusters
from greenheart.simulation.technologies.hydrogen.electrolysis.PEM_H2_LT_electrolyzer_Clusters import PEM_H2_Clusters as PEMClusters
# from PEM_H2_LT_electrolyzer_Clusters import PEM_H2_Clusters as PEMClusters
import numpy as np
from numpy import savetxt #ESG
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
import warnings
import math
import scipy
import time
from scipy import interpolate
#from PyOMO import ipOpt !! FOR SANJANA!!
warnings.filterwarnings("ignore")

"""
Perform a LCOH analysis for an offshore wind + Hydrogen PEM system

1. Offshore wind site locations and cost details (4 sites, $1300/kw capex + BOS cost which will come from Orbit Runs)~
2. Cost Scaling Based on Year (Have Weiser et. al report with cost scaling for fixed and floating tech, will implement)
3. Cost Scaling Based on Plant Size (Shields et. Al report)
4. Future Model Development Required:
- Floating Electrolyzer Platform
"""
# 
#---------------------------
# 
class run_PEM_clusters:
    '''Add description and stuff :)'''
    def __init__(self,electrical_power_signal,system_size_mw,num_clusters,verbose=True):

        self.cluster_cap_mw = np.round(system_size_mw/num_clusters)
        self.num_clusters = num_clusters

        self.stack_rating_kw = 1000
        self.stack_min_power_kw=0.1*self.stack_rating_kw
        #self.num_available_pem=interconnection_size_mw
        self.input_power_kw=electrical_power_signal
        self.cluster_min_power = self.stack_min_power_kw * self.cluster_cap_mw
        self.cluster_max_power = self.stack_rating_kw * self.cluster_cap_mw

        self.verbose = verbose

    def run(self):
        
        clusters = self.create_clusters()
        power_to_clusters = self.even_split_power()
        h2_df_ts = pd.DataFrame()
        h2_df_tot = pd.DataFrame()
        # h2_dict_ts={}
        # h2_dict_tot={}
        
        col_names = []
        start=time.perf_counter()
        for ci,cluster in enumerate(clusters):
            cl_name = 'Cluster #{}'.format(ci)
            col_names.append(cl_name)
            h2_ts,h2_tot = clusters[ci].run(power_to_clusters[ci])
            # h2_dict_ts['Cluster #{}'.format(ci)] = h2_ts
            
            h2_ts_temp = pd.Series(h2_ts,name = cl_name)
            h2_tot_temp = pd.Series(h2_tot,name = cl_name)
            if len(h2_df_tot) ==0:
                # h2_df_ts=pd.concat([h2_df_ts,h2_ts_temp],axis=0,ignore_index=False)
                h2_df_tot=pd.concat([h2_df_tot,h2_tot_temp],axis=0,ignore_index=False)
                h2_df_tot.columns = col_names

                h2_df_ts=pd.concat([h2_df_ts,h2_ts_temp],axis=0,ignore_index=False)
                h2_df_ts.columns = col_names
            else:
                # h2_df_ts = h2_df_ts.join(h2_ts_temp)
                h2_df_tot = h2_df_tot.join(h2_tot_temp)
                h2_df_tot.columns = col_names

                h2_df_ts = h2_df_ts.join(h2_ts_temp)
                h2_df_ts.columns = col_names

        end=time.perf_counter()
        if self.verbose:
            print('Took {} sec to run the RUN function'.format(round(end-start,3)))
        return h2_df_ts, h2_df_tot
        # return h2_dict_ts, h2_df_tot
    def optimize_power_split(self):
        #Inputs: power signal, number of stacks, cost of switching (assumed constant)
        #install PyOMO
        #!!! Insert Sanjana's Code !!!
        #
        power_per_stack = []
        return power_per_stack #size
    def even_split_power(self):
        start=time.perf_counter()
        #determine how much power to give each cluster
        num_clusters_on = np.floor(self.input_power_kw/self.cluster_min_power)
        num_clusters_on = np.where(num_clusters_on > self.num_clusters, self.num_clusters,num_clusters_on)
        power_per_cluster = [self.input_power_kw[ti]/num_clusters_on[ti] if num_clusters_on[ti] > 0 else 0 for ti, pwr in enumerate(self.input_power_kw)]
        
        power_per_to_active_clusters = np.array(power_per_cluster)
        power_to_clusters = np.zeros((len(self.input_power_kw),self.num_clusters))
        for i,cluster_power in enumerate(power_per_to_active_clusters):#np.arange(0,self.n_stacks,1):
            clusters_off = self.num_clusters - int(num_clusters_on[i])
            no_power = np.zeros(clusters_off)
            with_power = cluster_power * np.ones(int(num_clusters_on[i]))
            tot_power = np.concatenate((with_power,no_power))
            power_to_clusters[i] = tot_power

        # power_to_clusters = np.repeat([power_per_cluster],self.num_clusters,axis=0)
        end=time.perf_counter()
        if self.verbose:
            print('Took {} sec to run basic_split_power function'.format(round(end-start,3)))
        #rows are power, columns are stacks [300 x n_stacks]


        return np.transpose(power_to_clusters)
    def run_distributed_layout_power(self,wind_plant):
        #need floris configuration!
        x_load_percent = np.linspace(0.1,1.0,10)
        
        #ac2ac_transformer_eff=np.array([90.63, 93.91, 95.63, 96.56, 97.19, 97.50, 97.66, 97.66, 97.66, 97.50])
        ac2dc_rectification_eff=np.array([96.54, 98.12, 98.24, 98.6, 98.33, 98.03, 97.91, 97.43, 97.04, 96.687])/100
        dc2dc_rectification_eff=np.array([91.46, 95.16, 96.54, 97.13, 97.43, 97.61,97.61,97.73,97.67,97.61])/100
        rect_eff = ac2dc_rectification_eff*dc2dc_rectification_eff
        f=interpolate.interp1d(x_load_percent,rect_eff)
        start_idx = 0
        end_idx = 8760
        nTurbs = self.num_clusters
        power_turbines = np.zeros((nTurbs, 8760))
        power_to_clusters = np.zeros((8760,self.num_clusters))
        ac2dc_rated_power_kw = wind_plant.turb_rating
        
        power_turbines[:, start_idx:end_idx] = wind_plant._system_model.fi.get_turbine_powers().reshape((nTurbs, end_idx - start_idx))/1000
        power_to_clusters = (power_turbines)*(f(power_turbines/ac2dc_rated_power_kw))
        
        # power_farm *((100 - 12.83)/100) / 1000

        clusters = self.create_clusters()
       
        h2_df_ts = pd.DataFrame()
        h2_df_tot = pd.DataFrame()
        # h2_dict_ts={}
        # h2_dict_tot={}
        
        col_names = []
        start=time.perf_counter()
        for ci,cluster in enumerate(clusters):
            cl_name = 'Cluster #{}'.format(ci)
            col_names.append(cl_name)
            h2_ts,h2_tot = clusters[ci].run(power_to_clusters[ci])
            # h2_dict_ts['Cluster #{}'.format(ci)] = h2_ts
            
            h2_ts_temp = pd.Series(h2_ts,name = cl_name)
            h2_tot_temp = pd.Series(h2_tot,name = cl_name)
            if len(h2_df_tot) ==0:
                # h2_df_ts=pd.concat([h2_df_ts,h2_ts_temp],axis=0,ignore_index=False)
                h2_df_tot=pd.concat([h2_df_tot,h2_tot_temp],axis=0,ignore_index=False)
                h2_df_tot.columns = col_names

                h2_df_ts=pd.concat([h2_df_ts,h2_ts_temp],axis=0,ignore_index=False)
                h2_df_ts.columns = col_names
            else:
                # h2_df_ts = h2_df_ts.join(h2_ts_temp)
                h2_df_tot = h2_df_tot.join(h2_tot_temp)
                h2_df_tot.columns = col_names

                h2_df_ts = h2_df_ts.join(h2_ts_temp)
                h2_df_ts.columns = col_names

        end=time.perf_counter()
        if self.verbose:
            print('Took {} sec to run the distributed PEM case function'.format(round(end-start,3)))
        return h2_df_ts, h2_df_tot
        []
    
    def max_h2_cntrl(self):
        #run as many at lower power as possible
        []
    def min_deg_cntrl(self):
        #run as few as possible
        []
    def create_clusters(self):
        start=time.perf_counter()
        stacks=[]
        # TODO fix the power input - don't make it required!
        # in_dict={'dt':3600}
        for i in range(self.num_clusters):
            stacks.append(PEMClusters(cluster_size_mw = self.cluster_cap_mw))
        end=time.perf_counter()
        if self.verbose:
            print('Took {} sec to run the create clusters'.format(round(end-start,3)))
        return stacks


if __name__=="__main__":

    system_size_mw = 1000
    num_clusters = 20
    cluster_cap_mw = system_size_mw/num_clusters
    stack_rating_kw = 1000
    cluster_min_power_kw = 0.1*stack_rating_kw*cluster_cap_mw
    num_steps = 200
    power_rampup = np.arange(cluster_min_power_kw,system_size_mw*stack_rating_kw,cluster_min_power_kw)
    
    # power_rampup = np.linspace(cluster_min_power_kw,system_size_mw*1000,num_steps)
    power_rampdown = np.flip(power_rampup)
    power_in = np.concatenate((power_rampup,power_rampdown))
    pem=run_PEM_clusters(power_in,system_size_mw,num_clusters)

    h2_ts,h2_tot = pem.run()
    []


import os
import sys
sys.path.append('')
from dotenv import load_dotenv
import pandas as pd
import json
from hopp.simulation.technologies.sites import SiteInfo
from hopp.simulation.technologies.sites import flatirons_site as sample_site

from hopp.utilities.keys import set_developer_nrel_gov_key
# from plot_reopt_results import plot_reopt_results
# from run_reopt import run_reopt
from greenheart.to_organize.H2_Analysis.hopp_for_h2 import hopp_for_h2
from greenheart.to_organize.H2_Analysis.hopp_for_h2 import run_h2a as run_h2a #no h2a function
from greenheart.to_organize.H2_Analysis.simple_dispatch import SimpleDispatch
from greenheart.to_organize.H2_Analysis.simple_cash_annuals import simple_cash_annuals
import hopp.simulation.technologies.hydrogen.electrolysis.run_h2_PEM as run_h2_PEM
import numpy as np
import numpy_financial as npf
from lcoe.lcoe import lcoe as lcoe_calc
import matplotlib.pyplot as plt
import warnings
from pathlib import Path
warnings.filterwarnings("ignore")
from hopp.tools.resource import *
from hopp.tools.resource.resource_loader import site_details_creator

from greenheart.to_organize import hopp_tools_steel
import copy

from greenheart.to_organize import plot_results
from greenheart.to_organize import run_profast_for_hydrogen
#from greenheart.to_organize import run_profast_for_hydrogen Unsure if it is meant to be run_profast
from greenheart.to_organize.hopp_tools_steel import hoppDict
import yaml

ROOT_DIR = Path(__file__).parents[2]

"""
Landbased LCOH analysis
"""

#Set API key
load_dotenv()
NREL_API_KEY = os.getenv("NREL_API_KEY")
set_developer_nrel_gov_key('NREL_API_KEY')  # Set this key manually here if you are not setting it using the .env

#Step 1: User Inputs for scenario
save_hybrid_plant_yaml = True # hybrid_plant requires special processing of the SAM objects
save_model_input_yaml = True # saves the inputs for each model/major function
save_model_output_yaml = True # saves the outputs for each model/major function

resource_year = 2013
atb_years = [
            2025,
            2030,
            2035
            ]
policy = {
    'No Policy': {'Wind ITC': 0, 'Wind PTC': 0, "H2 PTC": 0},
    'Base': {'Wind ITC': 0, 'Wind PTC': 0.0051, "H2 PTC": 0.6},
    'Max': {'Wind ITC': 0, 'Wind PTC': 0.0256, "H2 PTC": 3},
}

sample_site['year'] = resource_year
useful_life = 25
critical_load_factor = 1
run_reopt_flag = False
custom_powercurve = True    #A flag that is applicable when using PySam WindPower (not FLORIS)
storage_used = False
battery_can_grid_charge = False
grid_connected_hopp = False
floris = False

# Technology sizing
interconnection_size_mw = 100
electrolyzer_size_mw = 100
wind_size_mw = 100
solar_size_mw = 0
storage_size_mw = 0
storage_size_mwh = 0

scenario_choice = 'Landbased Analysis'


# Site details
N_lat = 5 #50 # number of data points
N_lon = 5 #95
desired_lats = np.linspace(23.833504, 49.3556, N_lat)
desired_lons = np.linspace(-129.22923, -65.7146, N_lon)
load_resource_from_file = False
resource_dir = ROOT_DIR / "resource_files" 
sitelist_name = 'filtered_site_details_{}_lats_{}_lons_{}_resourceyear'.format(N_lat, N_lon, resource_year)

if load_resource_from_file:
    # Loads resource files in 'resource_files', finds nearest files to 'desired_lats' and 'desired_lons'
    site_details = resource_loader_file(resource_dir, desired_lats, desired_lons, resource_year)  # Return contains
    site_details.to_csv(os.path.join(resource_dir, 'site_details.csv'))
    site_details = filter_sites(site_details, location='usa only')
else:
    # Creates the site_details file containing grid of lats, lons, years, and wind and solar filenames (blank
    # - to force API resource download)
    if os.path.exists(sitelist_name):
        site_details = pd.read_csv(sitelist_name)
    else:
        site_details = site_details_creator.site_details_creator(desired_lats, desired_lons, resource_dir)
        # Filter to locations in USA
        site_details = filter_sites(site_details, location='usa only')
        site_details.to_csv(sitelist_name)
print("Site Details Created")

scenario = dict()
kw_continuous = electrolyzer_size_mw * 1000
load = [kw_continuous for x in
        range(0, 8760)]  # * (sin(x) + pi) Set desired/required load profile for plant

#Site lat and lon will be set by data loaded from Orbit runs

# Financial inputs
discount_rate = 0.0824
debt_equity_split = 0

# Wind costs input from ORBIT analysis
h2_model ='Simple'  #Basic cost model based on H2a and HFTO program record for PEM electrolysis
# h2_model = 'H2A'

# These inputs are not used in this analysis (no solar or storage)
solar_cost_kw = 9999
# storage_cost_kw = 250
# storage_cost_kwh = 240

# Flags (TODO: remove them and update documentation)
forced_sizes = True # no REopt
force_electrolyzer_cost = False


# Enable Ability to purchase/sell electricity to/from grid. Price Defined in $/kWh
# sell_price = 0.01
# buy_price = 0.01
sell_price = False
buy_price = False

# Set paths for results, floris and orbit
parent_path = os.path.abspath('')
results_dir = parent_path + '/examples/H2_Analysis/results/'
results = pd.DataFrame()
floris_dir = parent_path + '/floris_input_files/'

print('Parent path = ', parent_path)

# Site specific turbine information
path = ('examples/H2_Analysis/landbased_costs_ATB.xlsx')
xl = pd.ExcelFile(path)


# which plots to show
plot_power_production = True
plot_battery = True
plot_grid = True
plot_h2 = True
plot_desal = True
plot_wind = True
plot_hvdcpipe = True
plot_hvdcpipe_lcoh = True

for option in policy:
    for atb_year in atb_years:
        for i, site_deet in enumerate(site_details.iterrows()):
            if i == 0: continue
            else:
                site_deet = site_deet[1]
                print(site_deet)
            lat = site_deet['Lat']
            lon = site_deet['Lon']
            site_name = (str(lat)+","+str(lon))
            location_number = site_deet['site_nums']

            hopp_dict = hoppDict(save_model_input_yaml, save_model_output_yaml)
            sub_dict = {
                'policy': policy[option],
                'atb_year': atb_year,
                'site_location': site_name,
                'parent_path': parent_path,
                'load': load,
                'kw_continuous': kw_continuous,
                'sample_site': sample_site,
                'discount_rate': discount_rate,
                'forced_sizes': forced_sizes,
                'force_electrolyzer_cost': force_electrolyzer_cost,
                'wind_size': wind_size_mw,
                'solar_size': solar_size_mw,
                'storage_size_mw': storage_size_mw,
                'storage_size_mwh': storage_size_mwh,
                'solar_cost_kw': solar_cost_kw,
                # 'storage_cost_kw': storage_cost_kw,
                # 'storage_cost_kwh': storage_cost_kwh,
                'debt_equity_split': debt_equity_split,
                'useful_life': useful_life,
                'critical_load_factor': critical_load_factor,
                'run_reopt_flag': run_reopt_flag,
                'custom_powercurve': custom_powercurve,
                'storage_used': storage_used,
                'battery_can_grid_charge': battery_can_grid_charge,
                'grid_connected_hopp': grid_connected_hopp,
                'interconnection_size_mw': interconnection_size_mw,
                'electrolyzer_size_mw': electrolyzer_size_mw,
                'scenario':
                    {
                        'Useful Life': useful_life,
                        'Debt Equity': debt_equity_split,
                        'discount_rate': discount_rate,
                    },
                'sell_price': False,
                'buy_price': False,
                'h2_model': h2_model,
                'results_dir': results_dir,
                'scenario_choice': scenario_choice,
            }

            hopp_dict.add('Configuration', sub_dict)

            plot_dict = {
                'plot':
                    {
                        'plot_power_production': True,
                        'plot_battery': True,
                        'plot_grid': True,
                        'plot_h2': True,
                        'plot_desal': True,
                        'plot_wind': True,
                        'plot_hvdcpipe': True,
                        'plot_hvdcpipe_lcoh': True,
                    }
            }

            hopp_dict.add('Configuration', plot_dict)

            # set policy values
            hopp_dict, scenario, policy_option = hopp_tools_steel.set_policy_values(hopp_dict, scenario, policy, option)
            print(scenario['Wind PTC'])

            scenario_df = xl.parse()
            scenario_df.set_index(["Parameter"], inplace = True)

            site_df = scenario_df["Site"]
            site_df["Representative coordinates"] = site_name


            turbine_model = str(int(site_df['Turbine Rating']))+'MW'
            turbine_rating = site_df['Turbine Rating']
            print(turbine_model, "Turbine model")

            # set turbine values
            hopp_dict, scenario, nTurbs, floris_config = hopp_tools_steel.set_turbine_model(hopp_dict, turbine_model, scenario, parent_path,floris_dir, floris)

            scenario['Useful Life'] = useful_life

            # financials
            hopp_dict, scenario = hopp_tools_steel.set_financial_info(hopp_dict, scenario, debt_equity_split, discount_rate)

            # set electrolyzer information
            electrolysis_scale = 'Centralized'
            electrolyzer_replacement_scenario = 'Conservative'
            hopp_dict, electrolyzer_capex_kw,capex_ratio_dist, electrolyzer_energy_kWh_per_kg, time_between_replacement =  hopp_tools_steel.set_electrolyzer_info(hopp_dict,atb_year,electrolysis_scale,electrolyzer_replacement_scenario,turbine_rating,direct_coupling = False)
            # Extract Scenario Information from ORBIT Runs
            # Load Excel file of scenarios
            # OSW sites and cost file including turbines 8/16/2022

            # site info
            hopp_dict, site_df, sample_site = hopp_tools_steel.set_site_info(hopp_dict, site_df, sample_site)

            #fixed_or_floating_wind = site_df['Substructure technology']
            site = SiteInfo(sample_site, hub_height=scenario['Tower Height'])

            hopp_dict.add('Configuration', {'site': site})

            #Assign scenario cost details
            if atb_year == 2025:
                total_capex = site_df['2025 Wind CapEx ($/kW)']
                wind_om_cost_kw = site_df['2025 Wind OpEx ($/kW-yr)']
                storage_cost_kw = site_df['2025 Storage ($/kW)']
                storage_cost_kwh = site_df['2025 Storage ($/kWh)']
            if atb_year == 2030:
                total_capex = site_df['2030 Wind CapEx ($/kW)']
                wind_om_cost_kw = site_df['2030 Wind OpEx ($/kW-yr)']
                storage_cost_kw = site_df['2030 Storage ($/kW)']
                storage_cost_kwh = site_df['2030 Storage ($/kWh)']
            if atb_year == 2035:
                total_capex = site_df['2035 Wind CapEx ($/kW)']
                wind_om_cost_kw = site_df['2035 Wind OpEx ($/kW-yr)']
                storage_cost_kw = site_df['2035 Storage ($/kW)']
                storage_cost_kwh = site_df['2035 Storage ($/kWh)']


            capex_multiplier = site_df['CapEx Multiplier']
            wind_cost_kw = copy.deepcopy(total_capex) * capex_multiplier


            #Plot Wind Data to ensure offshore data is sound
            wind_data = site.wind_resource._data['data']
            wind_speed = [W[2] for W in wind_data]
            plot_results.plot_wind_results(wind_data, location_number, site_df['Representative coordinates'], results_dir, plot_wind)


            # Run HOPP

            hopp_dict, combined_hybrid_power_production_hopp, energy_shortfall_hopp, combined_hybrid_curtailment_hopp, hybrid_plant, wind_size_mw, solar_size_mw, lcoe = \
                hopp_tools_steel.run_HOPP(
                    hopp_dict,
                    scenario,
                    site,
                    sample_site,
                    forced_sizes,
                    solar_size_mw,
                    wind_size_mw,
                    storage_size_mw,
                    storage_size_mwh,
                    wind_cost_kw,
                    solar_cost_kw,
                    storage_cost_kw,
                    storage_cost_kwh,
                    kw_continuous,
                    load,
                    electrolyzer_size_mw,
                    wind_om_cost_kw,
                    nTurbs,
                    floris_config,
                    floris,
                )
            wind_plant_size = hybrid_plant.wind.system_capacity_kw
            print('Wind plant size: ',hybrid_plant.wind.system_capacity_kw)

            #Step 4: Plot HOPP Results
            plot_results.plot_HOPP(combined_hybrid_power_production_hopp,
                                    energy_shortfall_hopp,
                                    combined_hybrid_curtailment_hopp,
                                    load,
                                    results_dir,
                                    location_number,
                                    atb_year,
                                    turbine_model,
                                    hybrid_plant,
                                    plot_power_production)

            #Step 5: Run Simple Dispatch Model
            hopp_dict, combined_pv_wind_storage_power_production_hopp, battery_SOC, battery_used, excess_energy = \
                hopp_tools_steel.run_battery(
                    hopp_dict,
                    energy_shortfall_hopp,
                    combined_hybrid_curtailment_hopp,
                    combined_hybrid_power_production_hopp
                )

            plot_results.plot_battery_results(
                combined_hybrid_curtailment_hopp,
                energy_shortfall_hopp,
                combined_pv_wind_storage_power_production_hopp,
                combined_hybrid_power_production_hopp,
                battery_SOC,
                battery_used,
                results_dir,
                location_number,atb_year,turbine_model,
                load,
                plot_battery,
            )

            # grid information
            hopp_dict, cost_to_buy_from_grid, profit_from_selling_to_grid, energy_to_electrolyzer = hopp_tools_steel.grid(
                hopp_dict,
                combined_pv_wind_storage_power_production_hopp,
                sell_price,
                excess_energy,
                buy_price,
                kw_continuous,
                plot_grid,
            )

            #Step 6: Run the H2_PEM model
            h2_model = 'Simple'
            hopp_dict, H2_Results, H2A_Results, electrical_generation_timeseries = hopp_tools_steel.run_H2_PEM_sim(
                hopp_dict,
                hybrid_plant,
                energy_to_electrolyzer,
                scenario,
                wind_size_mw,
                solar_size_mw,
                electrolyzer_size_mw,
                kw_continuous,
                electrolyzer_capex_kw,
                lcoe,
            )

            plot_results.plot_h2_results(H2_Results,
                                        electrical_generation_timeseries,
                                        results_dir,
                                        location_number,atb_year,turbine_model,
                                        load,
                                        plot_h2)

            #Step 6b: Run desal model
            hopp_dict, desal_capex, desal_opex, desal_annuals = hopp_tools_steel.desal_model(
                hopp_dict,
                H2_Results,
                electrolyzer_size_mw,
                electrical_generation_timeseries,
                useful_life,
            )

            revised_renewable_cost = hybrid_plant.grid.total_installed_cost
            hydrogen_annual_production = H2_Results['hydrogen_annual_output']
            hydrogen_storage_capacity_kg = 0
            hydrogen_storage_cost_USDprkg = 0
            water_cost = 0.006868 #($/gal) average of green steel sites' water cost

            h2_ptc = scenario['H2 PTC']
            wind_ptc = scenario['Wind PTC']
            h2a_solution,h2a_summary,lcoh_breakdown,electrolyzer_installed_cost_kw = run_profast_for_hydrogen.run_profast_for_hydrogen(site_name,electrolyzer_size_mw,H2_Results,\
                            electrolyzer_capex_kw,time_between_replacement,hydrogen_storage_capacity_kg,hydrogen_storage_cost_USDprkg,\
                            desal_capex,desal_opex,useful_life,water_cost,wind_size_mw,solar_size_mw, \
                            hybrid_plant,revised_renewable_cost,wind_om_cost_kw,grid_connected_hopp,h2_ptc,wind_ptc)
            lcoh = h2a_solution['price']
            print('LCOH: ', lcoh)
            # # Max hydrogen production rate [kg/hr]
            max_hydrogen_production_rate_kg_hr = np.max(H2_Results['hydrogen_hourly_production'])
            max_hydrogen_delivery_rate_kg_hr  = np.mean(H2_Results['hydrogen_hourly_production'])

            electrolyzer_capacity_factor = H2_Results['cap_factor']

            print_results = False
            print_h2_results = False
            test = dict()

            test['Site Name'] = site_name
            test['Lat'] = lat
            test['Lon'] = lon
            test['ATB Year'] = atb_year
            test['Plant life'] = useful_life
            test['Policy'] = option
            test['Turbine size (MW)'] = turbine_rating
            test['Wind Plant size (MW)'] = wind_size_mw
            test['Wind Plant Size Adjusted for Turbine Rating(MW)'] = wind_plant_size /1000
            test['Electrolyzer size (MW)'] = electrolyzer_size_mw
            test['Load Profile (kW)'] = kw_continuous
            test['Energy to Electrolyzer (kW)'] = np.sum(energy_to_electrolyzer)
            test['Wind capacity factor (%)'] = hybrid_plant.wind.capacity_factor
            test['Electrolyzer capacity factor (%)'] = electrolyzer_capacity_factor
            test['LCOH ($/kg)'] = lcoh
            test['LCOH: Compression & storage ($/kg)'] = lcoh_breakdown['LCOH: Compression & storage ($/kg)']
            test['LCOH: Electrolyzer CAPEX ($/kg)']= lcoh_breakdown['LCOH: Electrolyzer CAPEX ($/kg)']
            test['LCOH: Desalination CAPEX ($/kg)'] = lcoh_breakdown['LCOH: Desalination CAPEX ($/kg)']
            test['LCOH: Electrolyzer FOM ($/kg)'] = lcoh_breakdown['LCOH: Electrolyzer FOM ($/kg)']
            test['LCOH: Electrolyzer VOM ($/kg)']=lcoh_breakdown['LCOH: Electrolyzer VOM ($/kg)']
            test['LCOH: Desalination FOM ($/kg)'] = lcoh_breakdown['LCOH: Desalination FOM ($/kg)']
            test['LCOH: Renewable plant ($/kg)'] = lcoh_breakdown['LCOH: Renewable plant ($/kg)']
            test['LCOH: Renewable FOM ($/kg)']= lcoh_breakdown['LCOH: Renewable FOM ($/kg)']
            test['LCOH: Taxes ($/kg)']=lcoh_breakdown['LCOH: Taxes ($/kg)']
            test['LCOH: Water consumption ($/kg)'] = lcoh_breakdown['LCOH: Water consumption ($/kg)']
            test['LCOH: Finances ($/kg)'] = lcoh_breakdown['LCOH: Finances ($/kg)']
            test['LCOH: total ($/kg)'] = lcoh_breakdown['LCOH: total ($/kg)']



            test = pd.DataFrame(test,index=[0])
            print(test)
            results = pd.concat([results,test])


results.to_csv(os.path.join(results_dir, "H2_Analysis_Landbased.csv"))
print('Done')


import os
import numpy as np
import pandas as pd
import scipy.optimize
from greenheart.simulation.technologies.hydrogen.electrolysis.PEM_tools import get_electrolyzer_BOL_efficiency

file_path = os.path.dirname(os.path.abspath(__file__))


def calc_efficiency_curve(operating_ratio, a, b, c, d):
    """Calculates efficiency [kWh/kg] given operation ratio with flattened end curves.

    Efficiency curve and general equation structure from Wang et. al (2023). See README.md
    in PEM_BOP directory for more details.

    Wang, X.; Star, A.G.; Ahluwalia, R.K. Performance of Polymer Electrolyte Membrane Water Electrolysis Systems: 
    Configuration, Stack Materials, Turndown and Efficiency. Energies 2023, 16, 4964. 
    https://doi.org/10.3390/en16134964

    Args:
        operating_ratio (list or np.array): Operation ratios.
        a (float): Coefficient a.
        b (float): Coefficient b.
        c (float): Coefficient c.
        d (float): Coefficient d.

    Returns:
        efficiency (list or np.array): Efficiency of electrolyzer BOP in kWh/kg.
    """
    efficiency = a + b * operating_ratio + c * operating_ratio**2 + d / operating_ratio

    return efficiency


def calc_efficiency(
    operating_ratio, efficiency, min_ratio, max_ratio, min_efficiency, max_efficiency
):
    """Adjust efficiency list to not go above minimum or maximum operating ratios in BOP_efficiency_BOL.csv

    Args:
        operating_ratio (list or np.array): Operation ratios.
        efficiency (list or np.array): Efficiencies calculated using curve fit.
        min_ratio (float): Minimum operating ratio from the CSV.
        max_ratio (float): Maximum operating ratio from the CSV.
        min_efficiency (float): Efficiency at the minimum operating ratio.
        max_efficiency (float): Efficiency at the maximum operating ratio.

    Returns:
        efficiency (list or np.array): Efficiencies limited with minimum and maximum values in kWh/kg.
    """
    efficiency = np.where(operating_ratio <= min_ratio, min_efficiency, efficiency)

    efficiency = np.where(operating_ratio >= max_ratio, max_efficiency, efficiency)
    return efficiency


def calc_curve_coefficients():
    """Calculates curve coefficients from BOP_efficiency_BOL.csv"""
    df = pd.read_csv(os.path.join(file_path, "BOP_efficiency_BOL.csv"))
    operating_ratios = df["operating_ratio"].values
    efficiency = df["efficiency"].values

    # Get min and max operating ratios
    min_ratio_idx = df["operating_ratio"].idxmin()  # Index of minimum operating ratio
    max_ratio_idx = df["operating_ratio"].idxmax()  # Index of maximum operating ratio

    # Get the efficiency at the min and max operating ratios
    min_efficiency = df["efficiency"].iloc[min_ratio_idx]
    max_efficiency = df["efficiency"].iloc[max_ratio_idx]

    # Get the actual min and max ratios
    min_ratio = df["operating_ratio"].iloc[min_ratio_idx]
    max_ratio = df["operating_ratio"].iloc[max_ratio_idx]

    curve_coeff, curve_cov = scipy.optimize.curve_fit(
        calc_efficiency_curve, operating_ratios, efficiency, p0=(1.0, 1.0, 1.0, 1.0)
    )
    return curve_coeff, min_ratio, max_ratio, min_efficiency, max_efficiency


def pem_bop(
    power_profile_to_electrolyzer_kw,
    electrolyzer_rated_mw,
    electrolyzer_turn_down_ratio,
):
    """
    Calculate PEM balance of plant energy consumption at the beginning-of-life
    based on power provided to the electrolyzer.

    Args:
        power_profile_to_electrolyzer_kw (list or np.array): Power profile to the electrolyzer in kW.
        electrolyzer_rated_mw (float): The rating of the PEM electrolyzer in MW.
        electrolyzer_turn_down_ratio (float): The electrolyzer turndown ratio.

    Returns:
        energy_consumption_bop_kwh (list or np.array): Energy consumed by electrolyzer BOP in kWh.
    """
    operating_ratios = power_profile_to_electrolyzer_kw / (electrolyzer_rated_mw * 1e3)

    curve_coeff, min_ratio, max_ratio, min_efficiency, max_efficiency = (
        calc_curve_coefficients()
    )

    efficiencies = calc_efficiency_curve(
        operating_ratios,
        *curve_coeff,
    )  # kwh/kg

    efficiencies = calc_efficiency(
        operating_ratios,
        efficiencies,
        min_ratio,
        max_ratio,
        min_efficiency,
        max_efficiency,
    )

    BOL_efficiency = get_electrolyzer_BOL_efficiency()  # kwh/kg

    BOL_kg = (electrolyzer_rated_mw * 1000) / BOL_efficiency  # kg/hr

    energy_consumption_bop_kwh = efficiencies * BOL_kg  # kwh

    energy_consumption_bop_kwh = np.where(
        power_profile_to_electrolyzer_kw
        < electrolyzer_turn_down_ratio * electrolyzer_rated_mw * 1000,
        0,
        energy_consumption_bop_kwh,
    )

    return energy_consumption_bop_kwh
